structure Env =

struct 		(* environments *)

   type env = (int * (int option)) list

   type item = env

   val emptyEnv = ([] : item)
   val null     = emptyEnv

   fun toString env = let

       fun doEntry (x,SOME a) = "x_"^(Int.toString x)^": "^(Int.toString a) 
 	|  doEntry (x,NULL)   = "x_"^(Int.toString x)^": Null"

       fun doit [] = "}"
	|  doit [p] = (doEntry p)^"}"
	|  doit (p::pp) = (doEntry p)^", "^(doit pp)

       in "{"^(doit env)
       end
       
   fun toWord [] = Word.fromInt 0
    |  toWord ((x,a) :: rest) = let
	val w = toWord rest
	val x = (Word.fromInt x) *  0wxd3
	val a = case a
		of NONE => 0wx7fffffff
		|  SOME a => Word.fromInt a
        in (w * 0wx1f + a + x)
	end

(* ... using environments ... *)

   fun lookup ((v,a)::rest) i = if v = i then a
				else lookup rest i
    |  lookup [] i            = (print "Error: unbound var. x_";
				 print ((Int.toString i)^"\n"); NONE)

   fun getValue (Horn.Const a) env = SOME a
    |  getValue (Horn.Var x)   env = lookup env x

   fun getFree (Horn.Const _) env = []
    |  getFree (Horn.Var x) env = if (lookup env x) = NONE then [x]
				  else []
   fun getFreePair (Horn.Const _, ind) env = getFree ind env
    |  getFreePair (ind, Horn.Const _) env = getFree ind env
    |  getFreePair (Horn.Var x, Horn.Var y) env      = if x = y then [x]
				    else [x,y]

   fun getFreeList args env = 
       IntUtils.remDup (List.concat (map (fn a => getFree a env) args))
       
   fun update [] x a = [(x, SOME a)]
    |  update ((x,a)::rest) y b = if x = y then (x,SOME b)::rest
   				  else (x,a)::(update rest y b)
    
   fun updatePair env (x1,x2) (a1,a2) =
       update (update env x1 a1) x2 a2

   fun updateList (env, [], [])       = env
    |  updateList (env, x::xx, a::aa) = let
       val env = update env x a
       in  updateList (env, xx, aa)
       end
    |  updateList (env, vars, aa) = (
				     print "\n\n ERROR in extending ";
				     print (toString env); print "\n\n";
				     env
				    )
   fun split [] env              = ([],[])	(* ok *)
    |  split (l as i::ii) env = (case getValue i env
       of SOME a => let
		val (p,s) = split ii env
		in (a::p,s)
		end
	| NONE   => ([],l)
	)

   fun unifyInd (env, Horn.Const a, b) = if a = b then SOME env 
					 else NONE
    |  unifyInd (env, Horn.Var x,   b) = let
       fun doit [] = (print "Warning! Occurrence of globally free var...\n";
		      SOME [(x, SOME b)])
	|  doit ((y,value)::rest) = if x = y then (case value
		of SOME a => if a = b then SOME env
			     else NONE
		|  NONE   => SOME ((x,SOME b)::rest)
		) else (case doit rest
		of SOME env => SOME ((y,value) :: env)
		|  NONE	    => NONE
		)
       in  doit env
       end	

   fun unify (env, [], []) = SOME env		(* ok *)
    |  unify (env, i::ii, a::aa) = (case getValue i env
	of SOME b => if a = b then unify (env,ii,aa)
		     else NONE
	|  NONE => let 
	   val Horn.Var x = i
	   val env = update env x a
	   in unify (env, ii, aa)
	   end
	)
    |  unify (env, [], _)  = (print "Error during unification: too few args\n"; NONE)
    |  unify (env, _, [])  = (print "Error during unification: too short result\n"; NONE)

   fun unifyAll (env, args, []) = []		(* ok *)
    |  unifyAll (env, args, h::t) = let
	  val result = unifyAll (env, args, t)
          in case unify (env, args, h)
	     of SOME env => env :: result
	      | NONE     => result
	  end

(* pooling env's in a forest ... *)

   fun envToIntList env = let
       fun f (_, SOME a) = a
	|  f (_, NONE) = ~1
       in map f env
       end

   fun insert (pool, p, env) = Forest.insert (pool, p, envToIntList env)
end 	(* end of Env *)
   
