structure HornEnvPool  =

struct

(* constraints *)

    structure F = Forest

    val envPools = ref (F.makeForest 1)
    fun resetTab () = envPools := (F.makeForest 1)
    
    type ind = int Horn.ind

   datatype  pre = 	U of int * ind list 
	        |	N  of int *  ind list 
		|	Eq of ind * ind
		|	Neq of ind * ind
                |       FunApp of ind * (int list -> int) * ind list     
		|	And of pre * pre
		|	Or of  pre * pre * int
		|	Exists of int * pre * int
		|	CheapOr of  pre * pre 
		|	CheapExists of int * pre 
		|	Yes | No
		
   datatype clause = 	One
		|	R of int * ind list 
		|	Implies of pre * clause
	        |       Break of clause
         	|	Both of clause * clause
		|	Forall of int * clause

   fun addPools clause = let

       fun doPre (Horn.U (r, args)) = U (r, args)
                                  
        |  doPre (Horn.N (r, args)) = N (r, args)
                                   
 	|  doPre (Horn.Eq (a1,a2)) = Eq (a1,a2)
	|  doPre (Horn.Neq (a1,a2)) = Neq (a1,a2)
        |  doPre (Horn.FunApp (y, f, args)) = FunApp(y, f, args)    
	|  doPre (Horn.And (p1,p2)) = (case (doPre p1, doPre p2)
		of (No,_) => No
		|  (_,No) => No
		|  (Yes,p) => p
		|  (p,Yes) => p
		|  (p1,p2) => And (p1,p2))
	|  doPre (Horn.Or (p1,p2)) = (case (doPre p1, doPre p2)
		of (Yes,_) => Yes
		|  (_,Yes) => Yes
		|  (No,p) => p
		|  (p,No) => p
		|  (p1,p2) => Or (p1,p2, Forest.new (!envPools)))
	|  doPre (Horn.Exists (x,p)) = Exists (x,doPre p, Forest.new (!envPools))
	|  doPre (Horn.CheapOr (p1,p2)) = (case (doPre p1, doPre p2)
		of (Yes,_) => Yes
		|  (_,Yes) => Yes
		|  (No,p) => p
		|  (p,No) => p
		|  (p1,p2) => CheapOr (p1,p2))
	|  doPre (Horn.CheapExists (x,p)) = CheapExists (x,doPre p)
	|  doPre Horn.Yes	= Yes
	|  doPre Horn.No	= No

       fun doClause Horn.One	= One
	|  doClause (Horn.R (x,args)) = R (x,args)
                                  
	|  doClause (Horn.Implies (p,c)) = (case doPre p
		of Yes => doClause c
		|  No  => One
		|  p   => Implies (p, doClause c))
        |  doClause (Horn.Break c) = Break (doClause c)
	|  doClause (Horn.Both (c1,c2))	 = (case (doClause c1, doClause c2)
		of (One, c) => c
		|  (c, One) => c
		|  (c1,c2)  => Both (c1,c2))
	|  doClause (Horn.Forall (x,c))  = (case doClause c
		of One => One
		|  c   => Forall (x,c))

       in doClause clause
       end
end




















