structure Solve =  

struct open Tail HornEnvPool

   structure F = Forest
   structure M = ForestMap

   type consumer = int list -> unit
   type infl = consumer list M.map

   type consumerN = int list * Env.env * (Env.env -> unit) 
   type consumerR = int * int list * Env.env * (Env.env list -> int list list)
   type consumerE = int -> unit
   type consumerNE = int * (int list -> int -> unit)
   
   type ncs = consumerN Stack.stack 
   type rcs = consumerR Stack.stack
   type eqcs = consumerE Stack.stack
   type neqcs = consumerNE Stack.stack 

   fun solve (result,clause) = let     
 
      val infl   = M.makeMap 1024 : infl 
      val ncs = Stack.new()
      val rcs = Stack.new()
      val eqcs = Stack.new()
      val neqcs = Stack.new()

      val count = ref 0 

      val clause = addPools clause

      fun assert r list = if F.insert (result, r, list) then let
          val rm = RelTable.getIdx r
	  fun doit (r,list, acc) = let        
        	val acc = case M.sub (infl, r, [])
			of SOME consumers => foldl (
				fn (f,acc) => (f, list) ::acc) 
				acc consumers 
			|  NONE => acc
		in case list 
		   of x::xs	=> let  
                                   val (_,j) = M.next (infl, r, x)
				   in doit (j, xs,acc)
				   end
	   	   |  []	=> acc 
		end
	  in app (fn (f,x) => f x) (doit (rm, list, [])) 
	  end
          else ()

      fun resumeEq (new) = if Stack.isEmpty(eqcs) then () else  
          let 
          val i = ref 0
          val len = Stack.length(eqcs)
          in while (!i < len) do let
          val f = Stack.sub(eqcs, !i)       
          val _ = i := !i + 1
          in f new
          end
      end

      fun resumeNeq (new) = if Stack.isEmpty(neqcs) then () else 
          let 
          val all = UniqueTree.getAll()
          val old = List.take (all, new)
          val i = ref 0
          val len = Stack.length(neqcs)
          in while (!i < len) do let
          val (id, f) = Stack.sub(neqcs, !i)       
          val _ = i := !i + 1
               
          in if (id = 0) then (f []) new 
                         else (app (f old) [new]; app (f [new]) old)
	  end
          end

      fun resumeN (new) = if Stack.isEmpty(ncs) then () else 
          let
          val all = UniqueTree.getAll()
          val old = List.take (all, new)

          val i = ref 0
          val len = Stack.length (ncs)
          in while (!i < len) do let 
             val (vars, env, checkFail) = Stack.sub(ncs, !i)
             val _ = i := !i + 1
     
             fun f x env = map (fn a => Env.update env x a) old
		 
             fun doit ([],    envList) = envList
	      |  doit (x::xs, envList) = doit (xs, 
			List.concat (map (f x) envList))
             val envFix = env
                 
             fun makeNew [] = []
               | makeNew (x::xs) = (Env.update envFix x new)::(makeNew xs)
                 
             fun filter [] = []
               | filter vars = let 
                       fun f [] a = [] 
                         | f (x::xs) a = if x<>a then x::(f xs a) else xs
                       val list = map (f vars) vars in list end
             
             fun newEnv ([], env) = env
               | newEnv (x::xs, env) = newEnv (xs, (Env.update env x new))       

             val newEnvList = makeNew vars 
             val varList = filter vars

             fun makeEnvList ([], _) = []
               | makeEnvList ((x::xs), (y::ys)) = doit(x, [y])@(makeEnvList (xs, ys))
                 
             val envList = (newEnv (vars, env)):: (makeEnvList (varList, newEnvList))
      in  app checkFail envList end
      end

      fun resumeR (new) = if Stack.isEmpty(rcs) then () else 
      let
      val all = UniqueTree.getAll()
      val old = List.take (all, new)

      val i = ref 0
      val len = Stack.length (rcs)
      in while (!i < len) do let 
         val (r, vars, env, mapL) = Stack.sub(rcs, !i)
         val _ = i := !i + 1             
         fun f x env = map (fn a => Env.update env x a) old
		
         fun doit ([],    envList) = envList
          |  doit (x::xs, envList) = doit (xs, List.concat (map (f x) envList))
         
         val envFix = env
                 
         fun makeNew [] = []
           | makeNew (x::xs) = (Env.update envFix x new)::(makeNew xs)
                 
         fun filter [] = []
           | filter vars = let 
                    fun f [] a = [] 
                      | f (x::xs) a = if x<>a then x::(f xs a) else xs
                    val list = map (f vars) vars in list end
                 
         fun newEnv ([], env) = env
           | newEnv (x::xs, env) = newEnv (xs, (Env.update env x new))       

         val newEnvList = makeNew vars 
         val varList = filter vars

         fun makeEnvList ([], _) = []
           | makeEnvList ((x::xs), (y::ys)) = doit(x, [y])@(makeEnvList (xs, ys))
                                  
         val envList = (newEnv (vars, env)):: (makeEnvList (varList, newEnvList))
         val lists = mapL envList
      in  app (assert r) lists end
      end

      fun check (pre, next) env = case pre

          of U (r,args) =>  let   
                val rm = RelTable.getIdx r
                val (prefix, rest) = Env.split args env        
		fun f t2 = case Env.unify (env,rest,t2)                 
			   of SOME env => next env   
	                   |  NONE => ()  
       		val list =  F.toList (result, r, prefix)
                   
       		val _ = if rest = [] andalso list <> [] then ()   
		        else M.add (infl, rm, prefix, f)        
		in app f list 
	   	end       

          |  N (r,args) => let                                         
		val vars = Env.getFreeList args env
                val all = UniqueTree.getAll()
                fun f x env = map (fn a => Env.update env x a) all
		fun doit ([],    envList) = envList
		 |  doit (x::xs, envList) = doit (xs, 
			List.concat (map (f x) envList)) 
		fun eval ([], env)    = []
                 |  eval (i::ii, env) = let
		    val SOME a = Env.getValue i env
		    in a :: (eval (ii, env))
		    end
                fun checkFail env = if F.hasElem (result, r, eval (args, env))
			then () else next env
                val envList = doit (vars, [env])
                val _ = if vars = [] then 1 else Stack.push (ncs, (vars, env, checkFail))
		in  app checkFail envList
		end     

          |  Eq (a1, a2)  => let
		 val b1 = Env.getValue a1 env
		 val b2 = Env.getValue a2 env
		 in case (b1,b2) 
		    of (SOME b1, SOME b2) => if b1 <> b2 then ()
			else next env
                    |  (SOME b1, NONE) => let 
			val Horn.Var x = a2
			in next (Env.update env x b1)
			end
                    |  (NONE, SOME b2) => let 
			val Horn.Var x = a1
			in next (Env.update env x b2)
			end
		    |  (NONE, NONE) => let              
                        val Horn.Var x1 = a1
                        val Horn.Var x2 = a2
                        in if x1=x2 then next env 
                        else let                             
                             val all = UniqueTree.getAll()
                             fun f a = next (Env.updatePair env (x1, x2) (a, a))
                             in (Stack.push (eqcs, f); app f all) 
		             end
                        end
                end

	  |  Neq (a1, a2) => let
                 val b1 = Env.getValue a1 env
                 val b2 = Env.getValue a2 env
                 val all = UniqueTree.getAll()
                 in case (b1,b2)
                    of (SOME b1, SOME b2) => if b1 = b2 then ()
                        else next env
                    |  (SOME b1, NONE) => let 
                        val Horn.Var x = a2
			fun f [] a = if a <>b1 then next (Env.update env x a)
				  else ()
			in (Stack.push (neqcs, (0,f)); app (f []) all)
        		end		
                    |  (NONE, SOME b2) => let
                        val Horn.Var x = a1
			fun f [] a = if a<>b2 then 
                                next (Env.update env x a)
				else ()
			in ( Stack.push (neqcs, (0, f)); app (f []) all)
                        end
                    |  (NONE, NONE) => let
                        val Horn.Var x1 = a1
                        val Horn.Var x2 = a2
                        in if x1 = x2 then ()
                        else let
                           fun f alist a = let
			       fun f1 b = if a<>b then next (
                                        Env.updatePair env (x1,x2) (a,b))
					else ()
                               in app f1 alist
                               end
			   in (Stack.push (neqcs, (1,f)); app (f all) all)
			   end
                        end
		  end

          | FunApp (y, ff, args) => let  
                fun eval ([], env) =[]
                 |  eval (i::ii, env) = let
                    val a = Env.getValue i env  
                    in case a of NONE => raise (Fail "\n\n\tSorry: unbound function argument!\n\n")
                              |  SOME v => v::(eval (ii, env)) 
		    end
                in let val args = eval(args, env)
                       val all = UniqueTree.getAll() 
		       val new = ff args
                       val _ = if (new <= List.last(all)) then () 
                               else (resumeR (new); resumeEq(new); 
				     resumeNeq(new); resumeN (new))
                   in case y 
			of Horn.Const i => if i = new then next env
				      else ()
                         | Horn.Var z   => next (Env.update env z new)
		   end
                end       
 
          |  And (pre1, pre2) => check (pre1, check (pre2,next)) env

          |  Or (pre1, pre2, pool) => let
		fun f env = if Env.insert (!envPools,pool,env) 
			    then next env else ()
                 val _ = check (pre1, f) env
                 in      check (pre2, f) env
                 end

          |  CheapOr (pre1, pre2) => let
                 val _ = check (pre1, next) env
                 in      check (pre2, next) env
                 end

          |  Exists (x, p, pool) => let
		fun f env = let 
		    val env = List.tl env 
                    in if Env.insert (!envPools,pool,env)
                       then next env
                       else ()
	            end
   		in check (p, f) ((x,NONE)::env)
		end

          |  CheapExists (x, p) =>    check (p, fn env => next (tl env)) ((x,NONE)::env)
          |  Yes => next env
          |  No =>  ()

      fun execute One env     = ()
       |  execute (R (r, args)) env  = let                                           
      		val vars = Env.getFreeList args env
                val all = UniqueTree.getAll()
                fun f x env = map (fn a => Env.update env x a) all
		fun doit ([],    envList) = envList
		 |  doit (x::xs, envList) = doit (xs, 
			List.concat (map (f x) envList)) 
		fun eval [] env    = []
                 |  eval (i::ii) env = let
		    val SOME a = Env.getValue i env
		    in a :: (eval ii env)
		    end
                fun mapL [] = []
                   | mapL (x::xs) = (eval args x)::(mapL xs)
                  
                val _ = if vars = [] then 1 else Stack.push (rcs, (r, vars, env, mapL)) 
                val envList = doit (vars, [env])
                val lists = mapL envList 
        	in  app (assert r) lists
		end  

       |  execute (Implies (pre, clause)) env =
                check (pre, execute clause) env
       |  execute (Break  clause) env = 
                ( count := (!count)+1
                ; execute clause env
                )
       |  execute (Both (c1, c2)) env = (
		execute c1 env;
                execute c2 env) 
       |  execute (Forall (x, clause)) env =
                execute clause ((x,NONE)::env)

   (* the actual constraint solving: *)

      val _ =  execute clause []
      val _ = print ("\nNumber of env's passing through breakpoint:\t" ^ Int.toString (!count) ^ "\n\n")

   in result
   end 	 	(* end of solve *)
end		(* end of Solve *)

















