structure Access =

struct
structure I = 
   struct open IntSets
   type item = IntSet
end

structure S = SymbolTable (structure Item = I)
structure B = BooleanMatrix 
structure T = Table (structure Item = I
		     type value = int)

fun getSuccs (set, []) = []
 |  getSuccs (set, a::aa) = if not (I.eqIntSets (set,a)) andalso
				    I.super (set,a) then 
			a :: getSuccs (set,aa)
			else getSuccs (set,aa)

(* ... TAKES a list of int sets and		*)
(*     RETURNS a matrix and a symbol table	*)
 
fun getMatrix list = let
    val table = S.makeTable 101
    fun f x = S.getIndex (table, x)

    val dom = IntUtils.remDup (map f list)

    fun doit i = (i,
		 map (fn x => (x,true)) (
		 IntUtils.remDup (
		 map f (
		 getSuccs (S.getItem (table,i), list))))
		)
    in (map doit dom, table)
    end	

fun getOrdering (table, path) = let
	fun f (next, last) = I.IntSet2List (
			     I.diffIntSets (next, last)) 

	fun doit [x] = [I.IntSet2List (S.getItem (table,x))]
	 |  doit [] = []
	 |  doit (x:: (xs as y::_)) = let
		val x = S.getItem (table,x)
		val y = S.getItem (table,y)
		in (f (x,y)) :: doit xs
		end

	val first = case path
		of [] => raise (Fail "\n\n\tSorry: empty path :-(\n\n")
		|  x::_ => S.getItem (table, x)

        in (first, List.concat (
	   rev (
	   doit path)))
	end

fun mapi f list = let
    fun doit (i,[]) = []
     |  doit (i, x::xs) = f (i,x) :: doit (i+1,xs)
    in doit (0,list)
    end

(* ... TAKES a maximal range of elements, and			*)
(*     a list of indices of int sets and			*)
(*     RETURNS a vetor of permutations together with		*)
(*     a table mapping int sets to indices of permutations 	*)
			
fun elab n list = let
	val max = I.IntList2Set (B.range n)

	val (matrix,table) = getMatrix list
	val paths = B.decompose matrix

	val result = T.makeTable 101

	fun f (i, path) = let
		val (first, list) = getOrdering (table, path)
		fun doit j = T.add (result, S.getItem (table,j), i)
		val _ = app doit path
		val rest = I.IntSet2List (I.diffIntSets (max, first))
		in list @ rest
		end
	val vec = Vector.fromList (mapi f paths)
	in (vec, result)
	end

val test = map I.IntList2Set [[1,3], [2,3,4], [3], [2,4], [2,3,4,5]]
end





