structure Reorder =

struct

   structure HQ = HornQ
   structure H  = Horn 
   structure I = 
	struct open IntSets
	   type item = IntSet
	end
   structure S = SymbolTable (structure Item = I)
   structure R = Table (structure Item = I
			type value = int)
   structure P = IntPairItem
   structure T = Table (structure Item = P
			type value = I.IntSet list)
   structure TT = Table (structure Item = P
			type value = int list Vector.vector * R.table)

   structure RT = RelTable


   fun permute (args, pi) = let
	fun f i = let
		fun doit (_,[]) = raise (Fail "\n\nIndex not found!\n\n")
		 |  doit (0,a::_) = a
		 |  doit (i,_::aa) = doit (i-1,aa)
		in doit (i,args)
		end
	in map f pi
	end

   (* fun rename newRel (n,i)  = let
		val (name,k) = RT.getItem n
		val newName = case i
			of ~1 => name
			|   0 => name ^ "-"
			|   _ => name ^ "-" ^ Int.toString i
		in RT.getIndex newRel (newName, k)
		end
   *)
      fun rename newRel (n,i, pi)  = let          (* Nov. 1 *)
		val (name,k) = RT.getItem n
		val newName = case i
		        of ~1 => name
			|   _ => let 
                                 fun doit ([], a, _) = a
                                   | doit (i::ii, a, b) = let 
                                     val c = if b then a else a^"-" 
                                     in doit (ii, c^(Int.toString i), true)
                                     end
		
		                 in doit (pi,name, false)
		                 end
		in RT.getIndex newRel (newName, k)
		end


   fun newQuery newRel (resTable, n,args,x) = 
	if I.isEmptyIntSet x then H.U (n, args)
	else let
	val k = List.length args
	val (vec,table) = case TT.sub (resTable, (n,k))
		of SOME (vec,table) => (vec,table)
		|  NONE => raise (Fail "\n\nRelation unknown :-(\n\n")
	val (j,pi) = case R.sub (table,x)
		of SOME j => (j,Vector.sub (vec,j))
		|  NONE   => raise (Fail "\n\nAccess set unknown :-(\n\n")
	val name = rename newRel (n,j, pi)       (* Nov. 1 *)
   	in H.U(name, permute (args,pi))
   	end

   fun auxClauses newRel (((name, k),(vec, table)), clause) = let
       fun forall list clause = foldr H.Forall clause list
       val vars = BooleanMatrix.range k
       fun doit (i,clause) = let

	val pi = Vector.sub (vec, i)
(*
		  handle _ => let
	val _ = print ("\n\nRelation:\t" ^ Int.toString name ^"/" ^ Int.toString k)
	val _ = print ("\nVector length:\t" ^ Int.toString (Vector.length vec))
	val _ = print ("\nIllegal index:\t" ^ Int.toString i ^ "\n\n")
	in raise (Fail "")
	end
*)

	val largs = map H.Var vars
	val rargs = map H.Var pi
	val newName = rename newRel (name,i, pi)              (* Nov.1 *)
	in H.Both (forall vars (H.Implies (	
		H.U(name,largs), H.R(newName,rargs))),clause)
	end
       in foldl doit clause (BooleanMatrix.range (Vector.length vec))
       end
	 
(* ... TAKES a mapping from names and arities to access set table	*)
(*     RETURNS a mapping to vectors and tables				*)

   fun elabAccess asTable = let
	fun f ((n,k), list) = ((n,k), Access.elab k list)
	in 	TT.fromList (
		map f (
		T.toList asTable))
	end

   fun allAuxClauses newRel (resTable, clause) = foldl (auxClauses newRel)
				 		   clause (TT.toList resTable)

   fun updateQueries newRel (resTable, clause) = let

	fun get ((n,k),set) = TT.sub (resTable, (n,k))

	fun doPre HQ.Yes = H.Yes
	 |  doPre HQ.No = H.No
	 |  doPre (HQ.CheapExists (int, pre)) = 
		   H.CheapExists (int, doPre pre)
	 |  doPre (HQ.Exists (int, pre)) =
		   H.Exists (int, doPre pre)
	 | doPre (HQ.CheapOr (p1,p2)) = 
		   H.CheapOr (doPre p1, doPre p2)
	 | doPre (HQ.Or (p1,p2)) = 
		   H.Or (doPre p1, doPre p2)
	 | doPre (HQ.And (p1,p2)) = 
		   H.And (doPre p1, doPre p2)
	 | doPre (HQ.Neq (x,y)) = H.Neq (x,y)
	 | doPre (HQ.Eq (x,y)) = H.Eq (x,y)
	 | doPre (HQ.FunApp (x,f,yy)) = H.FunApp (x,f,yy)
	 | doPre (HQ.N (n, args)) = H.N (n, args)
	 | doPre (HQ.U (n, args,x)) = newQuery newRel (resTable, n,args,x)

	fun doit HQ.One = H.One
	 |  doit (HQ.Forall (int,clause)) = H.Forall (int, doit clause)
	 |  doit (HQ.Both (c1,c2)) = H.Both (doit c1, doit c2)
	 |  doit (HQ.Implies (pre,c)) = 
		  H.Implies (doPre pre, doit c)
	 |  doit (HQ.Break c) = H.Break (doit c)
         |  doit (HQ.R (n, args)) = H.R (n, args)

	in doit clause
	end

(* ... putting all things together ...		*)

   fun reorder newRel clause = let
	val (clause, asTable) = HQ.elab clause
	val resTable = elabAccess asTable
	val clause = updateQueries newRel (resTable, clause)
	val clause = allAuxClauses newRel (resTable, clause)
	in clause 
	end
end


  
