structure Intern  =

struct open InternData

structure H = HornPlus

fun toString (Copy str) = str
 |  toString tree = let 
    val strList = ref ")\n"
    fun push str = (strList := str ^ (!strList))
    fun doit tree = (case tree 
        of Forall (str,t) => (push ") ";
			      doit t;
			      push ("Intern.Forall (\""^str^"\",\n\t\t ")
			     ) 
	|  Exists (str,t) => (push ")";
                              doit t;
                              push ("Intern.Exists (\""^str^"\",\n\t\t ")
                             )
	|  And (t1,t2) =>    (push ")";
                              doit t2;
			      push ",\n\t\t ";
			      doit t1;
			      push "Intern.And ("
			     )
	|  Or (t1,t2) =>    (push ")";
                              doit t2;
			      push ",\n\t\t ";
			      doit t1;
			      push "Intern.Or ("
			     )
	|  Implies (t1,t2) =>    (push ")";
                              doit t2;
			      push ",\n\t\t ";
			      doit t1;
			      push "Intern.Implies ("
			     )
        |  Break t =>    (
			      push "Intern.Break (";
			      doit t;
			      push ")\n\t\t "
			     )
	|  Yes	=> push "Intern.Yes"
	|  No   => push "Intern.No"
	|  Lit (str,tt) =>   (push "])";
			      doList tt;
			      push ("Intern.Lit (\""^str^"\",\n\t\t [")
			     )
	|  NLit (str,tt) =>  (push "])";
                              doList tt;
                              push ("Intern.NLit (\""^str^"\",\n\t\t [")
                             )
	|  Eq (t1,t2) =>     (push ")";
                              doit t2;
                              push ",\n\t\t ";
                              doit t1;
                              push "Intern.Eq ("
                             )
	|  Neq (t1,t2) =>     (push ")";
                              doit t2;
                              push ",\n\t\t ";
                              doit t1;
                              push "Intern.Neq ("
                             )
	|  Atom str	=>   push ("Intern.Atom \""^str^"\"")
	|  CopyString str	=>   push ("Intern.Atom ("^str^")")
	|  Copy str  =>    push str
       )
    and doList []      = ()
     |  doList [t]     = doit t
     |  doList (t::tt) = (
			  doList tt;
			  push ",\n\t\t ";
			  doit t
			)

    val _ = (doit tree; push " (")
    in (!strList)
    end


fun genClause t = let

    val vars = Vars.new ()
    val doTree = UniqueTree.intern 

    fun doArg (tree as Atom a) = if Vars.bound (vars,a)
		  then H.Var (Vars.get (vars,a))
		  else H.Const (doTree tree)
     |  doArg tree     = H.Const (doTree tree)

    val doArgs = map doArg

(* dealing with explicit unifications ... *)

    fun isGround (Atom str) = not (Vars.bound (vars, str))
     |  isGround (Lit (_, args)) = let
        fun doit [] = true
         |  doit (a::aa) = if isGround a then doit aa
		           else false
        in doit args
        end 
     |  isGround _ = raise (Fail "\tillegal tree ...")

    fun unify1 (t1, x) = case t1
        of Atom str => if Vars.bound (vars, str) 
            then H.Eq (H.Var (Vars.get (vars,str)), H.Var x)
	    else H.Eq (H.Const (doTree t1), H.Var x)
         | Lit (f, args) => if isGround t1 
            then H.Eq (H.Const (doTree t1), H.Var x)
            else let
                fun doit (pre, []) = H.U (f,(H.Var x)::List.rev pre)
                 |  doit (pre, t::tt) = if isGround t
                     then doit ((H.Const (doTree t))::pre, tt)
                     else (case t
                      of Atom str   =>
                              doit ((H.Var (Vars.get (vars, str)))::pre, tt)
                       | _ => let 
                              val y = Aux.newAux ()
                              val p = unify1(t,y)
                              val _ = Aux.popAux ()
                              in  H.CheapExists (y, H.And (p, doit ((H.Var y)::pre, tt)))
                             end)
                in  doit ([], args)
                end                          

     | _ => raise (Fail "\terror during unification ...")

    fun unify2 (x, t2) = case t2
        of Atom str => if Vars.bound (vars, str) 
            then H.Eq (H.Var x, H.Var (Vars.get (vars,str)))
            else H.Eq (H.Var x, H.Const (doTree t2))
        |  Lit (f, args) => if isGround t2 
            then H.Eq (H.Var x, H.Const (doTree t2))
            else let
                 fun doit (pre, []) = H.U (f, (H.Var x)::List.rev pre)
                  |  doit (pre, t::tt) = if isGround t
                     then doit ((H.Const (doTree t))::pre, tt)
                     else (case t
                      of Atom str   =>
                              doit ((H.Var (Vars.get (vars, str)))::pre, tt)
                       | _ => let 
                              val y = Aux.newAux ()
                              val p = unify2(y,t)
                              val _ = Aux.popAux ()
                              in  H.CheapExists (y, H.And (doit ((H.Var y)::pre, tt), p))
                              end)
                in  doit ([], args)
                end
        | _ => raise (Fail "\terror during unification ...")

    fun unify (t1,t2) = case t1
        of Atom str => if Vars.bound (vars, str) 
            then unify2 (Vars.get (vars, str), t2)
	    else (case t2
	          of Atom str' => if Vars.bound (vars, str') 
                      then H.Eq (H.Const (doTree t1), H.Var (Vars.get (vars, str')))
	              else if str = str' 
                           then H.Yes
	        	   else H.No
                  |  _ => H.No)
        |  Lit (f, args) => (case t2 
	       of Atom str => if Vars.bound (vars, str) 
                   then unify1 (t1, Vars.get (vars, str))
	           else H.No
		| Lit (f',args') => if f = f' 
                   then let
                        fun doit ([],[]) = H.Yes
		         |  doit ([],_) = H.No
		         |  doit (_,[]) = H.No
		         |  doit (a1::r1,a2::r2) =
			         H.And (unify (a1,a2), doit (r1,r2))
			in  doit (args, args') end
	           else H.No
		 | _ => raise (Fail "\tillegal term ..."))

        | _ => raise (Fail "\terror during unification ...")

(*	end of explicit unfications	*)

    fun doNeqLit (t, x) = let
        val auxPre = ref [] 
        fun genTerm (y,t) = case t 
        of Lit (f, args) => let
	    val args = doArgs args
	    in H.U (f, H.Var y :: args) end
	|  _ => raise Fail ("\n\nIllegal code for generating terms :-) ...\n\n")

	and doArgs [] = []
	 |  doArgs (t::aa) = (case t
	    of Atom str => if Vars.bound (vars, str) 
                then H.Var (Vars.get (vars,str)) :: doArgs aa
		else H.Const (doTree t) :: doArgs aa
             | Lit (f, args) => if isGround t 
                then H.Const (doTree t) :: doArgs aa
		else let 
                     val y = Aux.newAux ()
	             val _ = auxPre := (y, genTerm (y,t)) :: (!auxPre)
		     in H.Var y :: doArgs aa end 
         | _ =>  raise (Fail "\tIlligal args during doNeqLit ..."))

        val pre = if isGround t 
                  then H.Neq (x, H.Const (doTree t))
                  else
                    let
                       val _ = print ("Here we are ...\n") 
                       val y = Aux.newAux ()
             	       val _ = auxPre := (y, genTerm (y,t)) :: (!auxPre)
		       in H.Neq (H.Var y, x) end 

	fun wrap [] pre = pre
         |  wrap ((y,p) :: rest) pre = let
              val _ = Aux.popAux ()
	      in wrap rest (H.CheapExists (y, H.And (p, pre)))
	      end

	in wrap (!auxPre) pre
	end

    fun doNeq1 (t1, x) = case t1
        of Atom str => if Vars.bound (vars, str) 
            then H.Neq (H.Var (Vars.get (vars,str)), H.Var x)
	    else H.Neq (H.Const (doTree t1), H.Var x)
         | Lit (f, args) => doNeqLit(t1, H.Var x)
         | _ => raise (Fail "\terror during doNeq ...")

    fun doNeq (t1,t2) = case t1
        of Atom str => if Vars.bound (vars, str)
            then doNeq1 (t2, Vars.get (vars, str))
	    else (case t2
	 	  of Atom str' => if Vars.bound (vars, str')
		      then H.Neq (H.Const (doTree t1), H.Var (Vars.get (vars, str')))
		      else H.Neq(H.Const (doTree t1), H.Const (doTree t2))
                  |  _ => H.Yes)
        |  Lit (f, args) => (case t2 
            of Atom str => if Vars.bound (vars, str)
                then doNeq1 (t1, Vars.get (vars, str))
		else H.Yes
	     | Lit (f',args') => if f = f' 
                then let
		     fun doit ([],[]) = H.No
		      |  doit ([],_) = H.Yes
		      |  doit (_,[]) = H.Yes
		      |  doit (a1::r1,a2::r2) =
			  H.And (doNeq (a1,a2), doit (r1,r2))
	             in  doit (args, args') end
		else H.Yes
	     | _ => raise (Fail "\tillegal term ..."))

        | _ => raise (Fail "\terror during doNeq ...")
 
    fun doPreLit (str, args) = let

        val auxPre = ref []

	fun genTerm (y,t) = case t 
              of Lit (f, args) => let
		  val args = doArgs args
		  in H.U (f, H.Var y :: args)
		  end
	      | _ => raise Fail ("\n\nIllegal code for generating terms :-) ...\n\n")

	and doArgs [] = []
	 |  doArgs (t::aa) = (case t
	     of Atom str => if Vars.bound (vars, str) 
                 then H.Var (Vars.get (vars,str)) :: doArgs aa
		 else H.Const (doTree t) :: doArgs aa
              | Lit (f, args) => if isGround t 
                 then H.Const (doTree t) :: doArgs aa
		 else let 
                      val y = Aux.newAux ()
	              val _ = auxPre := (y, genTerm (y,t)) :: (!auxPre)
		     in H.Var y :: doArgs aa end
              | _ =>  raise (Fail "\tIlligal args during doPreLit ..."))

	val pre = H.U (str, doArgs args)

	fun wrap [] pre = pre
         |  wrap ((y,p) :: rest) pre = let
	      val _ = Aux.popAux ()
	      in wrap rest (H.CheapExists (y, H.And (pre,p)))
	      end

	in wrap (!auxPre) pre
	end

   fun doPreNLit (str, args) = let

        val auxPre = ref []

	fun genTerm (y,t) = case t 
        of  Lit (f, args) => let
	      val args = doArgs args
	      in H.U (f, H.Var y :: args)
	      end
	|  _ => raise Fail ("\n\nIllegal code for generating terms :-) ...\n\n")

	and doArgs [] = []
	 |  doArgs (t::aa) = (case t
	of Atom str => if Vars.bound (vars, str) 
            then H.Var (Vars.get (vars,str)) :: doArgs aa
	    else H.Const (doTree t) :: doArgs aa
         | Lit (f, args) => if isGround t 
            then H.Const (doTree t) :: doArgs aa
	    else let 
                 val y = Aux.newAux ()
	         val _ = auxPre := (y, genTerm (y,t)) :: (!auxPre)
		 in H.Var y :: doArgs aa
		 end
         | _ =>  raise (Fail "\tIlligal args during doPreNLit ..."))

	val pre = H.N (str, doArgs args)

	fun wrap [] pre = pre
         |  wrap ((y,p) :: rest) pre = let
	      val _ = Aux.popAux ()
	      in wrap rest (H.CheapExists (y, H.And (p,pre)))
	      end

	in wrap (!auxPre) pre
	end		
			  
	
    fun doPre p bool = case p 
	of Exists (str, p) => let
                val x = Vars.add (vars, str)
                val p = doPre p false
                val _ = Vars.remove (vars,str)
                in if bool then H.Exists (x,p)
                   else H.CheapExists (x,p)
                end
	|  And (p1,p2) => H.And (doPre p1 true, doPre p2 true)
	|  Or  (p1,p2) => if bool 
                          then H.Or (doPre p1 false, doPre p2 false)
			  else H.CheapOr (doPre p1 false, doPre p2 false)
	|  Yes => H.Yes
	|  No  => H.No
	|  Lit (str,args) => doPreLit (str, args)
	|  Eq (a1,a2) => unify (a1, a2)
        |  Neq (t1, t2) => doNeq (t1, t2)
        |  NLit (str,args) => doPreNLit (str, args)
	|  _ =>  raise (Fail "Pre-condition expected!\n\n")

    fun doLit (str, args) = let
        val auxClauses = ref []

	fun genTerm (y,t) = case t 
              of Lit (f, args) => let
	          val args = doArgs args
		  in (H.FunApp (H.Var y, UniqueTree.applyCons f, args),
		                H.R (f, H.Var y :: args))
		  end
	|  _ => raise Fail ("\n\nIllegal code for generating terms :-) ...\n\n")

	and doArgs [] = []
	  | doArgs (t::aa) = (case t
	      of Atom str => if Vars.bound (vars, str) 
                  then H.Var (Vars.get (vars,str)) :: doArgs aa
		  else H.Const (doTree t) :: doArgs aa
          | Lit (f, args) => if isGround t 
                  then H.Const (doTree t) :: doArgs aa
	 	  else let 
                       val y = Aux.newAux ()
                       val tempClause = (y, genTerm (y,t)) 
                       val _ = auxClauses :=(!auxClauses) @ [tempClause]
		       in H.Var y :: doArgs aa end
          | _ =>  raise (Fail "\tIlligal args during doLit ..."))

	val conclusion = H.R (str, doArgs args)

	fun wrap [] clause = clause
         |  wrap ((y,(pre,ass)) :: rest) clause = let
	     val _ = Aux.popAux ()
	     in H.Forall (y, H.Implies (pre, H.Both (ass, wrap rest clause)))
	     end

	in wrap (!auxClauses) conclusion
	end	
			  
    fun doClause t = case t
        of Forall (str, t) => let
             val x = Vars.add (vars, str)
             val c = doClause t
             val _ = Vars.remove (vars,str)
             in H.Forall (x,c) end
        |  And (t1, t2) => H.Both (doClause t1, doClause t2)
        |  Implies (p, t) => H.Implies (doPre p true, doClause t)
        |  Break t => H.Break (doClause t)
        |  Lit (str,args) => doLit (str,args)
        |  Yes  => H.One
        |  _    => raise (Fail "Clause expected!\n\n")

   in doClause t
   end
end









