structure Output =
struct 
    structure UT = UniqueTree
    structure RT = RelTable  
    structure Tab = HornEnvPool
   
   fun intListToCharList (list) = let
       fun doit ([], result) = #")" :: result
        |  doit ([x], result) = UT.toCharList0 (x,[] ) @ (#")" :: result)     
        |  doit (x::xs, result) = UT.toCharList0 (x,[]) @ (#"," :: #" " :: 
			doit (xs, result))
       in #"(" :: doit (list, []) end

   fun intListToString (list) = String.implode (intListToCharList (list))

   fun intListListToString (list) = let
       fun doit [] = []
        |  doit (first::rest) =  (intListToCharList (first)) ::
			      [#",", #" "] :: doit rest
       in String.implode (#"\t" :: List.concat (doit list))
       end

   fun printResult result = let
       val all = UT.getAll ()
       val _ = if all = [] then
                print "Empty universe, sorry!!!\n"
               else (print ("The Universe:\t");
                     print (intListToString (all));
                     print "\n\n")

       fun printRel (i,item) = let
	   fun doit list = (print (intListToString (list)); 
			    print ", ")
           in print ("Relation "^(RT.PairItem.toString item)^":\n\t");
	      Forest.appF doit (result,i,[]);  
              print "\n\n"
           end

       val _ = app printRel (RT.getList ())

       in print "\n" end

   fun countResult result = let
       val all = UT.getAll ()
       val _ = if all = [] 
               then print "Empty universe, sorry!!!\n"
               else (print ("The Universe:\t");
		     print (Int.toString (List.length all));
		     print "\n\n")

       fun printRel (i,item) = 
           (    print ("Relation "^(RT.PairItem.toString item)^":\t");
	        print (Int.toString (Forest.count (result,i)));
                print "\n"
           )
       val _ = app printRel (RT.getList ())
       in print "\n"
       end

   fun outputResult str result = let
       val out = TextIO.openOut str
       val all = UT.getAll ()
       val _ = if all = [] 
               then TextIO.output (out, "Empty universe, sorry!!!\n")
               else (TextIO.output (out, "The Universe:\t");
                     TextIO.output (out, intListToString (all));
                     TextIO.output (out, "\n\n")
                    )

       fun outputRel (i,item) = let
 	   fun doit list = (TextIO.output (out, intListToString (list));
			    TextIO.output (out, ", "))
           in
                TextIO.output (out, "Relation "^(RT.PairItem.toString item)^":\n\t");
                Forest.appF doit (result,i,[]);
                TextIO.output (out, "\n\n")
           end

       val _ = app outputRel (RT.getList ())

       in TextIO.output (out, "\n"); TextIO.closeOut out 
       end
    
   fun output str (select, univ) result = let
       val out = TextIO.openOut str
       val all = UT.getAll ()

       fun outputRel2 (i,item) = let
	   fun doit list = (TextIO.output (out, intListToString (list));
                            TextIO.output (out, ", "))
           in
                TextIO.output (out, "Relation "^(RT.PairItem.toString item)^":\n\t");
                Forest.appF doit (result,i,[]);
                TextIO.output (out, "\n\n")
	   end

       fun outputRel i = outputRel2 (i, RT.getItem i)

       val _ = if all = [] 
               then TextIO.output (out, "Empty universe, sorry!!!\n")
               else (case univ 
                     of true => (TextIO.output (out, "The Universe:\t");
                                 TextIO.output (out, intListToString (all));
                                 TextIO.output (out, "\n\n"))
                      | false => () )
       in case select
          of [] => let
		val _ = app outputRel2 (RT.getList ())
                in TextIO.output (out, "\n"); TextIO.closeOut out
                end
         |  list => let 
                fun f (r,k) = 
	            if RT.hasIndex (r,k) 
                    then outputRel2 (RT.getIndex (fn () => ~1) (r,k), (r,k))
	            else ()
	        in (app f list; TextIO.closeOut out)
                end
       end

 fun outputS (select, univ) result = let
       val out = TextIO.stdOut
       val all = UT.getAll ()

       fun outputRel2 (i,item) = let
	   fun doit list = (TextIO.output (out, intListToString (list));
                            TextIO.output (out, ", "))
           in
                TextIO.output (out, "Relation "^(RT.PairItem.toString item)^":\n\t");
                Forest.appF doit (result,i,[]);
                TextIO.output (out, "\n\n")
	   end

       val _ = if all = [] 
               then TextIO.output (out, "Empty universe, sorry!!!\n")
               else (case univ 
                     of true => (TextIO.output (out, "The Universe:\t");
                                 TextIO.output (out, intListToString (all));
                                 TextIO.output (out, "\n\n"))
                      | false => ())
       in case select
          of [] => let
		val _ = app outputRel2 (RT.getList ())
                in TextIO.output (out, "\n")
                end
         |  list => let fun f (r,k) = 
			     if RT.hasIndex (r,k) then
                             outputRel2 (RT.getIndex (fn () => ~1) (r,k), (r,k))
			     else ()
			 in (app f list)
			 end
       end

  fun outputStr (select, univ) result = let 
       val all = UT.getAll ()
     
       fun listToString [] = ""
        |  listToString (x::xs) = x ^", "^ listToString xs
 
       fun outputRel2 (i,item) =
           let val d = listToString (map intListToString (Forest.toList (result, i, [])))
               val itemStr = RT.PairItem.toString item 
           in ("Relation "^itemStr^":\n\t"^d^"\n\n") end

       val ustr = if all = [] 
                  then "Empty universe, sorry!!!\n"
                  else (case univ 
                        of true => "The Universe:\t"^ (intListToString (all))^"\n\n"
                        | false => "" )               

       in case select
          of [] => ustr ^ (listToString (map outputRel2 (RT.getList ())))
          |  list => let fun f (r, k) = 
			     if RT.hasIndex (r,k) then
                             outputRel2 (RT.getIndex (fn () => ~1) (r,k), (r,k))
			     else ""
                        in ustr ^ (listToString(map f list))
			end
       end

  fun outputD (select, univ) result = let 

       val all = UT.getAll ()

       fun outputRel2 (i,item) =
           let val d = map (map UT.toString) (Forest.toList (result, i, []))
               val itemStr = RT.PairItem.toString item 
           in (itemStr, d) end

       val ulist = if all = [] 
                   then (print "Empty universe, sorry!!!\n"; [])
                   else (case univ 
                         of true => map UT.toString all
                          | false => [] )                    
       in case select
          of [] => (ulist, map outputRel2 (RT.getList ()))
         |  list => let fun f (r, k) = 
			     if RT.hasIndex (r,k) then
                             outputRel2 (RT.getIndex (fn () => ~1) (r,k), (r,k))
			     else (r, [])
                        in (ulist, map f list)
			end
       end

end















