structure TransList =

struct open HornPlus

    structure UT = UniqueTree 
    structure RT = RelTable
 
    fun translate newRel clauses = let
        val aux = ref (Horn.One : (int,int) Horn.clause)                      

        fun toFacts t = let
             val (f,aa) = UT.T.getItem (!UT.unique,t)
	     in case aa
	        of NONE      => ()
                 | SOME args => let
	            val k = List.length args
		    val r = RT.getIndex newRel (f,k+1)
		    in case args
        	       of [] => (aux := Horn.Both (Horn.R (r,[Horn.Const t]),!aux)) 
        	       | args => let
           			 fun doit [] = (aux := Horn.Both (Horn.R (r, 
                                                map Horn.Const (t::args)), !aux))
            		 	   | doit (a::aa) = (toFacts a; doit aa)
           			  in  doit args	end
	            end
             end			
      
        fun transInd (Var x) = Horn.Var x
         |  transInd (Const t) = let 
	        val _ = toFacts t
		in Horn.Const t
		end

       fun transIndList list = map transInd list

       fun transPre (U (r, args))    = let 
	        val k = List.length args
		val i = RT.getIndex newRel (r,k)
		in Horn.U (i, transIndList args)
		end
        |  transPre (N (r, args))    = let 
		val k = List.length args
		val i = RT.getIndex newRel (r,k)
		in Horn.N (i, transIndList args)
		end

        |  transPre (Eq (a,b))	     = Horn.Eq (transInd a, transInd b)
        |  transPre (Neq (a,b))	     = Horn.Neq (transInd a, transInd b)
	|  transPre (FunApp (a,f,bb))= Horn.FunApp (transInd a, f, transIndList bb)
	|  transPre (And (p1,p2))    = Horn.And (transPre p1, transPre p2)
	|  transPre (Or (p1,p2))     = Horn.Or (transPre p1, transPre p2)
	|  transPre (CheapOr (p1,p2))     = Horn.CheapOr (transPre p1, transPre p2)
	|  transPre (Exists (x,p))   = Horn.Exists (x, transPre p)
	|  transPre (CheapExists (x,p))   = Horn.CheapExists (x, transPre p)
 	|  transPre Yes		     = Horn.Yes
	|  transPre No		     = Horn.No

       fun transClause (R (r, args))   = let 
                val k = List.length args
                val i = RT.getIndex newRel (r,k)
		in Horn.R (i, transIndList args)
		end
	|  transClause (Both (c1,c2))  = Horn.Both (transClause c1, transClause c2)
	|  transClause (Forall (x,c))  = Horn.Forall (x, transClause c)
	|  transClause (Implies (p,c))  = Horn.Implies (transPre p, transClause c)
        |  transClause (Break c)  = Horn.Break (transClause c)
	|  transClause One	       = Horn.One

       val clauses = map transClause clauses
       val clause = let 
		val t:: rest = clauses
		val t = Horn.Both (!aux,t)
		in case rest
		   of [] => Reorder.reorder newRel t
		   |  _  => foldr Horn.Both t rest
		end
       in  clause
       end
end










