structure UniqueTree = struct 

structure P = Pair (structure First  = StringItem structure Second = IntListOptionItem) 
structure T = SymbolTable (structure Item = P)

type item = int
val unique = ref (T.makeTable 1024)
val all = ref []: int list ref

fun resetTrees () = (unique := (T.makeTable 1024); all := [])

fun applyCons cname intlist = let 
    val idx = T.getIndex (!unique, (cname, SOME intlist))
    val _ =  if null(!all) then all:= [idx] 
             else if idx <= List.last(!all) then () 
                  else all := (!all) @ [idx] 
    in idx end
   
fun internTree tree = case tree
    of Tree.Atom a => T.getIndex (!unique, (a,NONE))
    |  Tree.Struct (f,args) => let	
       val args = map internTree args
       val idx = T.getIndex (!unique, (f,SOME args))
       val _ = if null(!all) then all:= [idx] 
               else if idx <= List.last(!all) then () 
                    else all := (!all) @ [idx]	
       in idx end   

fun intern tree = case tree
    of InternData.Lit (f,args) => let	
       val args = map intern args
       val idx = T.getIndex (!unique, (f, SOME args))
       val _ = if null(!all) then all:= [idx] 
               else if idx <= List.last(!all) then () 
                    else all := (!all) @ [idx] 
       in idx end	
    |  InternData.Atom a =>  let 
       val idx = T.getIndex (!unique, (a,NONE))
       val _ = if null(!all) then all:= [idx] 
               else if idx <= List.last(!all) then () 
                    else all := (!all) @ [idx] 
       in idx end
   |  _ =>  raise (Fail "\tIllegal trees during intern tree ...")

fun toWord int = Word.fromInt int   

fun toCharList0 (tree,acc) = let	
    val (f,aa) = T.getItem (!unique,tree)
	
    fun doArgs ([],acc) = #")" :: acc 
      | doArgs ([a],acc) = toCharList0 (a, #")" :: acc)
      | doArgs (a::aa,acc) = toCharList0 (a, #"," :: #" " :: doArgs (aa,acc))
	
    in case aa
       of NONE => String.explode f @ acc
	| SOME args => String.explode f @
			    #" " :: #"(" :: doArgs (args,acc)
    end

fun toString tree = String.implode (toCharList0 (tree,[]))

fun getAll () = !all
end
