structure ParseClause (* : ParseClause *) =
   struct
      structure LrVals = ClauseLrValsFun(structure Token = LrParser.Token)
      structure Lex = ParserLexFun(structure Tokens = LrVals.Tokens)
      structure Parser = Join(structure ParserData = LrVals.ParserData
			      structure Lex=Lex
			      structure LrParser=LrParser)
	 
      fun printError fname (s,(l1,c1),(l2,c2)) = 
	 TextIO.output(TextIO.stdErr,
		       String.concat["[",fname,
				     ":",Int.toString l1,".",Int.toString c1,
				     "-",Int.toString l2,".",Int.toString c2,
				     "] ",s,"\n"
				     ])

      fun genLexer fname = let val f = TextIO.openIn fname
			       val i = fn x => TextIO.inputN(f,x)
			   in (f,Parser.makeLexer i)
			   end

      fun parseClause fname = 
	 let 
	    val (f,lexer) = genLexer fname
	    val (interns,_) = Parser.parse(15,lexer,printError fname, ())
	    val _ = TextIO.closeIn f
	 in 
	    map Intern.genClause interns
	 end

(* Used for constraints from string *)

      fun genLexerStr str = let val id = ref 0
                                 val i = fn x => if (!id = 0) then (let val _ = id := !id +1 in str end) else "\n"                  
			   in Parser.makeLexer i 
			   end		
   
      fun parseClauseStr str = 
	 let 
	    val lexer = genLexerStr str
	    val (interns,_) = Parser.parse(15,lexer,printError "error", ())
	   (*  val _ = TextIO.closeIn f *)
	 in 
	    map Intern.genClause interns
	 end 
   (* used for parsing constraints from standard input April 9 *)

      fun genLexerIN () = let val f = TextIO.stdIn
			      val i = fn x => TextIO.inputN(f, 1)
			   in Parser.makeLexer i
			   end
			   
      fun parseClauseIN () = 
	 let 
	    val lexer = genLexerIN () 
	    val (interns,_) = Parser.parse(15,lexer,printError "error", ())
	   (*  val _ = TextIO.closeIn f *)
	 in 
	    map Intern.genClause interns
	 end

    (* This is added by Sun on Mar 25, 2003 for output intern datastructure *)
     fun parseOutput fname =
         let 
	    val (f,lexer) = genLexer fname
	    val (interns,_) = Parser.parse(15,lexer,printError fname, ())
	    val _ = TextIO.closeIn f
	 in 
            interns
	 end

   end
