functor FrontBackEnd (structure ParseClause : ParseClause) =
struct 
structure A = AnalyzeTreeList
structure O = Output

(* initialization *)
fun init () = let 
    val result = Forest.makeForest 1
    val _ = UniqueTree.resetTrees () 
    val _ = RelTable.resetTab ()
    val _ = HornEnvPool.resetTab ()
    in result end

fun init1 (result) = let 
    val _ = Forest.resetForest (result)
    val _ = UniqueTree.resetTrees () 
    val _ = RelTable.resetTab ()
    val _ = HornEnvPool.resetTab ()
    in () end 

(* input clause from HornDirect data structure *)
fun inputData (clause) = let 
    val cl = Intern.genClause (HornDirect.translate clause)
    in [cl] end

(* input clause from file *)
fun inputFile (fileName) = ParseClause.parseClause fileName  

(* input clause from StdIn *)
fun inputStd () = ParseClause.parseClauseIN () 

(* input clause from string *) 
fun inputStr (str) = ParseClause.parseClauseStr str

(* output the selected result to string *)
fun outputData result (select, univ) = O.outputD (select, univ) result
 
(* output the selected result to file *)
fun outputFile (result, fileName) (select, univ) = O.output (fileName) (select, univ) result

(* output the selected result to stdOut *)
fun outputStd result (select, univ) = O.outputS (select, univ) result

(* output the selected result to string *)
fun outputStr result (select, univ) = O.outputStr (select, univ) result


(* solve the result without output *) 
fun solve (result, cl) = A.onlySolve (result, cl)

(* solve and count the result without output *) 
fun solveCount (result, cl) = A.solveAndCount (result, cl)

(* solve and output to  SML data structure *)
fun solveData (result, cl) = A.outputAllD (result, cl)

(* solve and output to file *)
fun solveFile (result, cl, fileName) = A.outputAll (fileName) (result, cl)

(* solve and output to stdOut *)
fun solveStd (result, cl) = A.outputAllS (result, cl)

(* solve and output to string *)
fun solveStr (result, cl) = A.outputAllStr (result, cl)

(* solve with selected output to SML data structure *)
fun selectSolveData (result, cl) (select, univ, count) = A.outputD (select, univ, count) (result, cl)

(* solve with selected output to file *)
fun selectSolveFile (result, cl, fileName) (select, univ, count) = 
          A.output (fileName) (select, univ, count) (result, cl)

(* solve with selected output to stdOut *)
fun selectSolveStd (result, cl) (select, univ, count) = A.outputS (select, univ, count) (result, cl)

(* solve with selected output to string *)
fun selectSolveStr (result, cl) (select, univ, count) = A.outputStr (select, univ, count) (result, cl)

(* transform the ALFP formula from file to the HornDirect clause *)
fun fromALFPtoHD (fileName) = map InternToDirect.toHdClause (ParseClause.parseOutput fileName)
                      
end
