(* This is to transform the clause in HornDirect to internD, Dec. 12 *)

structure  HornDirect = 

struct

   structure I = InternData

   datatype term =	Cons of string	 
                |       Var of string
                |       AppF of string * term list    
 
   datatype pre = 	Prel of string * term list
		|	Nrel of string * term list
		|	Pconj of pre list
		|	Pdisj of pre list
		|	Exists of string list * pre
		|       Eq of term * term
                |       Neq of term * term
                |       true 
                |       false

		
   datatype clause = 	One 
		|	Rel of string * term list
		|	Impl of pre * clause
	        |	BImpl of pre * clause
		|	Conj of clause list
		|	Forall of string list * clause

   

   fun translate clauses = let
      
       fun transTerm (Cons t) = I.Atom t
        |  transTerm (Var t) = I.Atom t
        |  transTerm (AppF (f, args)) = I.Lit (f, map transTerm args) 

       fun transPre (Prel (r, args))  = I.Lit (r, map transTerm args) 
        |  transPre (Nrel (r, args))  = I.NLit (r, map transTerm args)
	|  transPre (Pconj plist)     = (case plist of 
           [] => I.Yes
         | [p1] => transPre p1
         | p1::ps => I.And (transPre p1, transPre (Pconj ps)))
       	|  transPre (Pdisj plist)     = (case plist of 
           [] => I.Yes
         | [p1] => transPre p1
         | p1::ps => I.Or (transPre p1, transPre (Pdisj ps)))
	|  transPre (Exists (xlist,p))= (case xlist of 
           [] => transPre p
         | x1::xs => I.Exists (x1, transPre (Exists (xs, p)))) 
        |  transPre (Eq (a,b))	      = I.Eq (transTerm a, transTerm b)
        |  transPre (Neq (a,b))       = I.Neq (transTerm a, transTerm b)
        |  transPre true = I.Yes
        |  transPre false = I.No

       fun transClause (Rel (r, args)) = I.Lit (r, map transTerm args)
	|  transClause (Conj clist)  = (case clist of 
           [] => I.Yes
         | [c1] => transClause c1
         | c1::cs => I.And (transClause c1, transClause (Conj cs)))
	|  transClause (Forall (xlist,c)) = (case xlist of 
           [] => transClause c
         | x1::xs => I.Forall (x1, transClause (Forall (xs, c))))
	|  transClause (Impl (p,c))  = I.Implies (transPre p, transClause c)
        |  transClause (BImpl (p,c)) = I.Implies (transPre p, I.Break(transClause c))
	|  transClause One	     = I.Yes

       in  transClause clauses end
end


