(* transform from intern to HornDirect *)
structure InternToDirect  =

struct open InternData

structure H = HornDirect

fun toHdClause t = let
    
    val vars = Vars.new ()

    fun toTerm t = case t of
	 Atom str => if Vars.bound (vars, str) then
	     	     H.Var str
		else H.Cons str
         | Lit (f, args) => H.AppF (f, map toTerm args)

    fun toHdPre p = case p 
	of 
	   Exists (str, p) => let   
                val x = Vars.add (vars, str)
                val pp = toHdPre p 
                val _ = Vars.remove (vars, str)
                in H.Exists ([str], pp)
                end
	|  And (p1,p2) => H.Pconj ([toHdPre p1]@[toHdPre p2])
	|  Or  (p1,p2) => H.Pdisj ([toHdPre p1]@[toHdPre p2])
	|  Yes => H.true 
	|  No  => H.false  
	|  Lit (str, args) => H.Prel (str, map toTerm args)
	|  Eq (a1,a2) => H.Eq (toTerm a1, toTerm a2)
        |  Neq (t1, t2) => H.Neq (toTerm t1, toTerm t2)
        |  NLit (str,args) => H.Nrel (str, map toTerm args)
	|  _ =>  raise (Fail "Pre-condition expected!\n\n")

			  
    fun toHdClause t = case t
        of Forall (str, t) => let
                val x = Vars.add(vars, str)
                val c = toHdClause t
                val _ = Vars.remove (vars, str)
                in H.Forall ([str], c)
                end
        |  And (t1, t2) => H.Conj ([toHdClause t1]@[toHdClause t2])
        |  Implies (p, t) => (case t of 
                              Break tt => H.BImpl(toHdPre p, toHdClause t)
                             | _       => H.Impl (toHdPre p, toHdClause t))
        |  Break t => toHdClause t
        |  Lit (str,args) => H.Rel (str, map toTerm args)
        |  Yes          => H.One
        |  _            => raise (Fail "Clause expected!\n\n")
  
   in toHdClause t
   end

(* get HornDirect data structure *)


end









