(* This is to print the ALFP clause from the HornDirect data structure into a Latex file *)

structure Pretty =
struct 
     structure H = HornDirect

fun StrListToString [] = ""
  | StrListToString [x] = x
  | StrListToString (x::xs) = x ^ "," ^ StrListToString xs

fun TermToString (H.Cons t) = "\\ALFPconst{" ^ t ^ "}"
  | TermToString (H.Var t) =  "\\ALFPconst{" ^ t ^ "}"
  | TermToString (H.AppF(f, args)) = "\\ALFPrel{" ^ f ^ "(" ^ TermListToString args ^ ")}" 
and
    TermListToString [] = ""
  | TermListToString [x] = TermToString x
  | TermListToString (x::xs) = (TermToString x) ^ "," ^ (TermListToString xs)

fun PreToString (H.Prel (r, args)) = "\\ALFPrel{" ^ r ^ "(" ^ TermListToString args ^")}"
  | PreToString (H.Nrel (r, args)) = "\\ALFPnot{\\ALFPrel{" ^ r ^ "(" ^ TermListToString args ^")}}" 
  | PreToString (H.Pconj (pl)) = (case pl of 
           [] => ""
         | [p] => PreToString p
         | p::ps =>  PreToString p ^ "\\ALFPconj" ^ PreToString (H.Pconj ps))
  | PreToString (H.Pdisj (pl)) =(case pl of 
           [] => ""
         | [p] => PreToString p
         | p::ps => PreToString p ^ "\\ALFPdisj" ^ PreToString (H.Pdisj ps))
  | PreToString (H.Exists(xl, p)) = "\\ALFPexist{" ^ StrListToString xl ^"}" ^ PreToString p
  | PreToString (H.Eq (a,b)) =  TermToString a ^ "=" ^ TermToString b
  | PreToString (H.Neq (a,b))  = TermToString a ^ "\\ALFPneq" ^ TermToString b
  | PreToString (H.true) = "\\textbf 1"
  | PreToString (H.false) = "\\textbf 0"
 
fun ClauseToString (H.One) = "\\textbf 1"
  | ClauseToString (H.Rel(r, args)) ="\\ALFPrel{" ^ r ^ "(" ^ TermListToString args ^")}"
  | ClauseToString (H.Impl(p, c)) =PreToString p ^ "\\ALFPimply" ^ ClauseToString c 
  | ClauseToString (H.BImpl(p, c)) =PreToString p ^ "\\ALFPlimply" ^ ClauseToString c 
  | ClauseToString (H.Conj(cl)) = (case cl of 
           [] => ""
         | [c] => ClauseToString c
         | c::cs => ClauseToString c ^ "\\ALFPconj \\\\\n" ^ ClauseToString (H.Conj cs))
  | ClauseToString (H.Forall(xl, c)) = "\\ALFPall{"^StrListToString xl^"}" ^ ClauseToString c


fun prettyTable (fname, cl, capt) = let  (* as a Table *)
    val fileout = TextIO.openOut(fname ^".tex")
    val _       = TextIO.output(fileout,
                                "\\begin{table}\n\\framebox[\\textwidth]{\n" ^
                                "\\begin{minipage}{\\textwidth}\n")
    val _       = TextIO.output(fileout, ClauseToString cl)
    val _       = TextIO.output(fileout, 
                                "\\end{minipage}}\n\\caption{" ^ capt ^ "}\n\\label{" ^ fname ^ "}\n" ^
                                "\\end{table}\n")
    in            TextIO.closeOut fileout 
    end

fun prettyFrame (fname, cl) = let   (* with frame *)
    val fileout = TextIO.openOut(fname ^".tex")
    val _       = TextIO.output(fileout,
                                "\\framebox[\\textwidth]{\n" ^
                                "\\begin{minipage}{\\textwidth}\n")
    val _       = TextIO.output(fileout, ClauseToString cl)
    val _       = TextIO.output(fileout, 
                                "\\end{minipage}}\n")
    in            TextIO.closeOut fileout 
    end

fun prettyText (fname, cl) = let  (* pure text *)
    val fileout = TextIO.openOut(fname ^".tex")
    val _       = TextIO.output(fileout, ClauseToString cl)
    in            TextIO.closeOut fileout 
    end

end
