functor Matrix (structure Content : Content) : Matrix =

struct  
	infix 4 ==

	structure C = Content
	structure I = Pair (structure First = IntItem
				    structure Second = IntItem)

	type content = C.content
	type matrix = (int * (int * content) list) list

	structure O =
	struct 
	type element = int
	fun cmp (x,y)= case Int.compare (x,y)
		of EQUAL => EQUAL
	 	|  GREATER => LESS
		|  LESS => GREATER
	end

	structure IS = Sort (structure E = O)

	structure Ordered_Entry =
	struct
	type element = int * int * content
	fun cmp ((x,y,v),(x',y',v')) = case Int.compare (x,x')
		of LESS 	=> LESS
		|  GREATER	=> GREATER
		|  EQUAL	=> Int.compare (y,y')
	end

	structure S = Sort (structure E = Ordered_Entry)

(* Outputting to Lists and Strings *)

        fun matrixToList m = let
	    fun f x (y, v) = (x,y,v)
	    fun doit (x,e) = map (f x) e
	    in List.concat (map doit m) 
            end

	fun getDomain m = let
	    fun doit [] = []
	     |  doit ((y,v)::e) = y::doit e
	    fun doX (x,e) = x :: doit e
	    in IS.remDup (List.concat (map doX m))
	    end

       fun getSuccs ([],x) = []
	|  getSuccs ((y,e)::m,x) = (case Int.compare (x,y)
		of GREATER	=>	getSuccs (m,x)
		|  EQUAL	=>	e
		|  LESS		=>	[]
	)

       fun getLine (m, x) = let
	   
	   val dom = getDomain m

	   fun doit ([],_) = []
	    |  doit (y::ys,[]) = (x,y,C.zero) :: doit (ys,[])
	    |  doit (y::ys, e as (z,v)::e') = (case Int.compare (y,z)
			of LESS		=> 	(x,y,C.zero) :: doit (ys,e)
			|  GREATER	=> 	(TextIO.print (
				Int.toString y); TextIO.print " vs. ";TextIO.print (
				Int.toString z); TextIO.print " in ";
				TextIO.print (PrintUtils.intListToString dom);
				raise (Fail "\n\tillegal successor :-(")
				)
			|  EQUAL	=> 	(x,y,v) :: doit (ys,e')
			)
           
	   fun doit1 m	= case m 
	      of []		=> 	raise (Fail ("\n\tillegal line index " ^ Int.toString x ^ " :-("))
	      | (z,e)::m 	=>	(case Int.compare (x,z) 
			of GREATER => doit1 m
			|  EQUAL=> (
(*
			TextIO.print ("\n" ^ Int.toString (List.length e));
*)
			doit (dom,e)
			)
			|  LESS => 	raise (Fail ("\n\tillegal line index " ^ Int.toString x ^ " :-(("))
			)
	   in doit1 m
	   end
	       
       fun lineToString (m, x) = let
	   fun f (_,_,v) = "\t"^C.toString v
	   in (String.concat (map f (getLine (m,x))))^"\n"
	   end
	       
       fun toString m = let

	   val dom = getDomain m
	   fun f x = lineToString (m,x)
	   val list = map f dom 
	   in String.concat list ^ "\n"
	   end

(* Creating matrices *)

        fun range k = let
            fun doit j = if j >= k then []
                else j :: doit (j+1)
            in doit 0
            end

        fun const list v = let
		val list = IS.remDup list
		fun f x = (x, map (fn y => (y,v)) list)
		in map f list : matrix
		end

	fun emptyMatrix () = [] : matrix

        fun zero list = let
	    val list = IS.remDup list
	    in map (fn x => (x,[])) list : matrix
	    end

        fun listToMatrix edges = let

	    val edges = S.sort edges
            val dom = let
		fun doit [] = []
		 |  doit ((x,y,_)::rest) = x::y:: doit rest
		in IS.remDup (doit edges)
		end

	    fun acc x (all as (z,y,v)::list, succs) = if x = z then
				            if v <> C.zero then acc x (list, (y,v) :: succs)
			           	    else                acc x (list, succs)
				else (all, (x, succs))
	     |  acc x ([],succs) = ([],(x,  succs))

	    fun doit ([],_,m) = m
	     |  doit (x::xs,edges, m) = (case edges 
		of []	=>	doit (xs, [], (x,[])::m)
		|  (z,_,_)::e => if x <> z then doit (xs, edges, (x,[])::m)
				 else let
				 val (edges, next) = acc x (edges, [])
				 in  doit (xs, edges, next::m)
				 end
		)
	    in doit (List.rev dom, edges,[])
	    end

	fun llToMatrix list = let
	    val k = List.length list

	    fun doit1 i values = let
	        fun doit (j,[]) = []
		 |  doit (j,v::vs) = if j >= k then []
			else if v <> C.zero 
			     then (j,v) :: doit (j+1,vs)
			     else          doit (j+1,vs)
		in (i, doit (0,values))
		end
  
	    fun doit2 (j,[]) = []
	     |  doit2 (j,v::vs) = if j >=k then []
			else (doit1 j v) :: doit2 (j+1,vs)

	    in doit2 (0,list)
	    end
	    
(* Calculating with matrices *)

	fun (m1 == m2) = ((matrixToList m1) = (matrixToList m2))

	fun merge ([],r) = r
	 |  merge (r,[]) = r
	 |  merge (r1 as (x,v)::r,r2 as (y,v')::r') = (case Int.compare (x,y)
		of EQUAL	=>	let
					val u = C.add (v,v')
					in if u = C.zero then merge (r,r')
					   else (x,u) :: merge (r,r')
					end
		 | LESS		=>	(x,v)  :: merge (r,r2)
		 | GREATER	=>	(y,v') :: merge (r1,r')
		)

        fun add (l1, l2) = case (l1, l2) 
	       of ([],[])			=> []
	       |  ((x,_)::rest, [])	=> (x,[]):: add (rest, [])
	       |  ([], (x,_)::rest)	=> (x,[]):: add ([], rest)
	       |  ((x,e)::r,(y,e')::r')	=> (case Int.compare (x,y)
			of EQUAL =>	(x, merge (e,e')) :: add (r, r')
			|  LESS  =>	(x,[]) :: add (r, l2) 
			|  GREATER =>	(y,[]) :: add (l1, r')
			)

       fun mergeList list       = let
           fun mergePairs []    = []
            |  mergePairs [l]   = [l]
            |  mergePairs (x::y::t) = (merge (x,y)):: mergePairs t
           fun doit list =  case list
                of []           => []
                |  [l]          => l
                |  _            => doit (mergePairs list)
           in doit list
           end                        

	fun mult (m1, m2) = let
	    
	     fun f (v, []) = []
	      |  f (v,(y,u)::rest) = let 
				val u = C.mult (u,v)
				in if u = C.zero then f (v, rest)
				   else (y, u) :: f (v, rest)
				end

		
	    fun vm (x,row) = let
		fun doit ([],m) = []
		 | doit (_,[]) = []
		 |  doit (vec as (y,v)::rest, m as (z,row):: m') = (case Int.compare (y,z)
			of LESS 	=> doit (rest, m)
			|  GREATER	=> doit (vec, m')
			|  EQUAL	=> (f (v, row)) :: doit (vec, m')
			)

		in (x, mergeList (doit (row, m2)))
		end

	    in map vm m1
	    end

	fun rectify (dom,m) = let
	    fun doit ([],_) = []
	     |  doit (x::xs, []) = (x,[]) :: doit (xs,[])
	     |  doit (x::xs, all as (y,row)::rest) = (case Int.compare (x,y) 
			of LESS		=> (x,[]) :: doit (xs, all)
			|  EQUAL	=> (y,row) :: doit (xs, rest)
			|  GREATER	=> (y,row) :: doit (x::xs, rest)
		)
	   in doit (dom,m)
	   end

	fun transpose m = let
	   val dom = getDomain m
	   fun f (x,y,v) = (y,x,v)
	   val list = matrixToList m
	   val list = map f list
	   val m = listToMatrix list
	   in rectify (dom,m)
	   end

        fun rem (m,x) = let
	    val rem2 = List.filter (fn (y,_) => y <> x)
	    fun f ((z,e), list) = if z = x then list
			else (z, rem2 e)::list
	    in foldr f [] m
	    end

	fun sub (m, (x,y)) = let
	    fun subE [] = C.zero
	     |  subE ((z,v)::e) = if z = y then v
		else subE e

	    in case m
		of []		=> C.zero
		|  (z,e)::m	=> if x = z then subE e
				   else sub (m,(x,y))
	    end

	fun map f m = let
	    fun doLine (x,row) = (x, List.map (fn (y,v) => (y, f v)) row)
	    in List.map doLine m
	    end


end 	(* end of structure Matrix *)
