structure Tokens = Tokens

open Tokens 

exception EofInComment
exception LexNotInteger

fun String2Const s = 
   case Int.fromString s
     of NONE => raise LexNotInteger
      | SOME n => n

type pos = int*int
type svalue = Tokens.svalue
type ('a,'b) token = ('a,'b) Tokens.token
type lexresult = (svalue,pos) token

val line = ref 1
val col = ref 0

fun incCol n = col := !col+n
fun incLine() = (line := !line+1; col := 0)
fun initLine() = (line := 1; col := 0)

fun pos() = (!line,!col)
fun pos2 n = ((incCol 1; pos()), (incCol (n-1); pos()))
fun pos3 x n = (x,(incCol 1; pos()),(incCol (n-1); pos()))

val eof = fn () => Tokens.EOF(pos(),pos())

(*
\n => (incLine(); lex());	
*)
%%

%header (functor ParserLexFun(structure Tokens: Clause_TOKENS));
%full
%s Copy Normal Comment;

sep    = [\ \t]; 
letter = [0-9a-zA-Z_'];

%%
<INITIAL>{sep}* => (YYBEGIN Copy; initLine(); lex());


<Copy> "<|" => 	(YYBEGIN Normal; (OPENFORMULA (pos2 2)));
<Copy> "||" =>  (YYBEGIN Normal; (CLOSELIT (pos2 2)));
<Copy> "}}" =>  (YYBEGIN Normal; (CLOSESTRING (pos2 2)));
<Copy> \n    =>  ((COPY (pos3 "\n" 1)) before incLine());
<Copy> .   =>  (COPY (pos3 yytext (String.size yytext)));

<Comment> "*)" => (YYBEGIN Normal; incCol 2; lex());
<Comment> \n => (incLine(); lex());	
<Comment> .    => (incCol 1; lex());

<Normal> {sep}* => (incCol (String.size yytext); lex());
<Normal> \n => (incLine(); lex());	
<Normal> "(*"   => (YYBEGIN Comment; incCol 2; lex());
<Normal> "||"   => (YYBEGIN Copy; OPENLIT (pos2 2));
<Normal> "{{"   => (YYBEGIN Copy; OPENSTRING (pos2 2));
<Normal> "|>"   => (YYBEGIN Copy; CLOSEFORMULA (pos2 2));

<Normal> E  =>   (EXISTS (pos2 1));
<Normal> A  =>   (FORALL (pos2 1));
<Normal> 1  =>   (YES (pos2 1));
<Normal> 0  =>  (NO (pos2 2));

<Normal> {letter}+   => (NAME(pos3 yytext (String.size yytext)));

<Normal> "("  => (OPEN(pos2 1));
<Normal> ")"  => (CLOSE(pos2 1)); 
<Normal> ","  => (COMMA(pos2 1));
<Normal> "!"  => (NOT(pos2 1)); 
<Normal> "."  => (DOT(pos2 1));
<Normal> "="  => (EQ(pos2 1));
<Normal> "!=" => (NEQ(pos2 2)); 
<Normal> "=>" => (IMPLIES(pos2 2));
<Normal> "&"  => (AND(pos2 1));
<Normal> "|"  => (OR(pos2 1));

