functor BinomialQueue(structure E : ORDERED_ELEMENT) : PRIORITY_QUEUE = 
struct
(*
 val cumTime   = ref(Time.zeroTime)
 fun addTime tr = let val t = Timer.checkCPUTimer tr
                  in
                    cumTime := Time.+(!cumTime, #usr t)
                  end

  val numOfOps  = ref 0
  val numOfCmps = ref 0
  fun inc x     = (x := !x + 1)
  fun cmp(a, b) = (inc numOfCmps; E.cmp(a, b))
  fun show()    = ( print "BinQueue : ";
                    print (real(!numOfCmps) / real(!numOfOps));
                    print "    ";
                    let val dev = open_out "bench.out"
                    in
                       output(dev, Time.toString(!cumTime));
                       close_out dev
                    end;
                    Basic.tell("PQ operations", makestring(!numOfOps));
                    Basic.tell("comparison"   , makestring(!numOfCmps));
                    Basic.tell("ratio", makestring(real(!numOfCmps) / real(!numOfOps))) )
*)

structure Element = E

exception Empty

local 
  datatype bin_tree  = Node of Element.element * int * bin_tree list

  fun elem(Node(e, _, _)) = e

  fun link(n1 as Node(r1, d1, children1), n2 as Node(r2, d2, children2)) = 
        if Element.cmp(r1, r2) = LESS 
          then Node(r1, d1+1, n2 :: children1)
          else Node(r2, d2+1, n1 :: children2)
     
  fun meld([], queue) = queue
   |  meld(queue, []) = queue
   |  meld((n1 as Node(r1, d1, _))::queue1, (n2 as Node(r2, d2, _))::queue2) = 
        if d1 < d2 then n1 :: meld(queue1, n2 :: queue2)
        else
        if d1 > d2 then n2 :: meld(n1 :: queue1, queue2)
        else meldCarry(queue1, queue2, link(n1,n2))          
  and meldCarry([], [], carry)    = [carry]
   |  meldCarry([], queue, carry) = meld(queue, [carry])
   |  meldCarry(queue, [], carry) = meld(queue, [carry])
   |  meldCarry( (n1 as Node(r1, d1, _)) :: queue1, 
                 (n2 as Node(r2, d2, _)) :: queue2, n3 as Node(r3, d3, _)) = 
       if d1=d2 
         then if d3 < d1
                then n3 :: meld(n1 :: queue1, n2 :: queue2)
                else n1 :: meldCarry(queue1, queue2, link(n2, n3))
         else if d1=d3
                then meldCarry(queue1, n2 :: queue2, link(n1, n3))
              else 
              if d2=d3
                then meldCarry(n1 :: queue1, queue2, link(n2, n3))
                else n3 :: meld(n1 :: queue1, n2 :: queue2)
  and meld1([], carry) = [carry]
   |  meld1((n1 as Node(r1, d1, _)) :: queue1, n2 as Node(r2, d2, _)) = 
        if d1 < d2 then n1 :: meld1(queue1, n2)
        else 
        if d1 > d2 then n2 :: n1 :: queue1
        else meld1(queue1, link(n1, n2))

  fun findMinimal((n as Node(r, _, _)) :: ns, i, min, minIndex) =
        if Element.cmp(r, elem min) = LESS
          then findMinimal(ns, i+1, n, i)
          else findMinimal(ns, i+1, min, minIndex)
   |  findMinimal([], _, min, minIndex) = (min, minIndex)

in
  type pqueue = bin_tree list

  val empty = []

  fun isEmpty([]) = true
   |  isEmpty(_)  = false

  fun dequeue []      = raise Empty
   |  dequeue (n::ns) =
        let (*val tr = Timer.startCPUTimer()*)
            fun deleteMin(x :: xs, index) =
                  if index = 0 then xs else x :: deleteMin(xs, index-1)
             |  deleteMin(_, _) = ErrorMsg.bug("BinomialQueue", "deleteMin")
            val (Node(r, _, children), minIndex) = findMinimal(ns, 1, n, 0)
        in
          let val e = (r, meld(rev children, deleteMin(n::ns, minIndex)))
          in
            (*addTime tr; inc numOfOps;*)
            e
          end
        end

  fun enqueue(queue, x) = 
         let (*val tr = Timer.startCPUTimer()*)
             fun f([], _, x) = [x]
              |  f((n as Node(_, d, _)) :: ns, i, x) = 
                    if i<d then x :: n :: ns
                           else f(ns, i+1, link(n, x))
         in
           let val e = f(queue, 0, Node(x, 0, []))
           in
              (* addTime tr; inc numOfOps; *)
              e
           end
        end

  fun findMin([])      = raise Empty
   |  findMin(n :: ns) = 
        let (* val tr = Timer.startCPUTimer() *)
            val Node(e, _, _) = #1(findMinimal(ns, 1, n, 0))
        in
           (* addTime tr; inc numOfOps; *)
           e
        end

  fun deleteMin queue = 
        let (* val tr = Timer.startCPUTimer() *)
            val e  = #2(dequeue queue)
        in
          (* addTime tr; inc numOfOps; *)
          e
        end
end (* LOCAL *)
end (* BinomialQueue *)
