signature ExtArray =
sig
   type 'a extArray
   val isEmpty: 'a extArray -> bool
   val sub :	'a extArray * int -> 'a
   val update :	'a extArray * int * 'a -> unit
   val modify    : ('a -> 'a) -> 'a extArray -> unit
   val tabulate : int * 'a * (int -> 'a) -> 'a extArray
   val array    : int * 'a -> 'a extArray
   val foldl    : ('a * 'b -> 'b) -> 'b -> 'a extArray -> 'b 
   val length : 'a extArray -> int
   val toList : 'a extArray -> 'a list
   val fromList: 'a -> 'a list -> 'a extArray
end

structure ExtArray :> ExtArray =
struct
   type 'a extArray = ('a * int * 'a Array.array) ref

   fun array (n, zero) = ref (zero, 0, Array.array (n,zero))

   fun isEmpty (ref (_,n,array))  = if n = 0 then true 
				  else false

   fun length (ref (_,n,array)) = n

   fun sub (ref (a,n,array) : 'a extArray,i)     = if i >= n then a 
				else Array.sub (array,i)

   fun update (r,i,a) = let
		val ref (zero,n,array) = r
		in if i < n then
                 Array.update (array,i,a)
              else let
                 fun copy j = if j < n then
                      Array.sub (array,j)
                      else if j = i then a
		      else zero
                 val newLength = if i >= 2 * n then i + 1
				 else 2 * n
                 in r := (zero, i+1, Array.tabulate (newLength, copy))
                 end
          end


   fun fromList zero list = let
       val array = Array.fromList list
       in
	   ref (zero,List.length list, array)
       end

   fun toList (ref (_,n,array)) = let
       fun doit (m, list) = if m >= n then List.rev list
		else let
		   val a = Array.sub (array,m)
                in doit (m+1,a::list)
                end
       in doit (0, [])
       end

   fun modify f  (ref (_,_, array)) = Array.modify f array

   fun foldl f b (ref (_,_, array)) = Array.foldl f b array

   fun tabulate (n, zero, f) = ref (zero,n, Array.tabulate (n,f))
end

   
	     
