structure Forest :> Forest =

struct
   structure T = Table (structure Item = IntPairItem
			    type value = int )

   type forest = int list option Stack.stack * T.table 
   
   fun max (stack,table) = T.max table
   fun zeroFraction (stack,table) = T.zeroFraction table
   fun average (stack,table) = T.average table

   fun makeForest n = (Stack.makeStack (n,NONE), T.makeTable n) : forest
   fun nullForest () = (Stack.new (), T.nullTable ())         : forest

   fun resetForest (stack, table) = (Stack.resetStack (stack); Stack.push(stack, NONE);
                                         T.resetTable (table,1)) 

   fun numberOfNodes (stack,_) = Stack.length stack

   fun initSucc ((stack,_), j, list) = case Stack.sub (stack, j)
	of SOME _	=>	false
	 | NONE		=>	(Stack.update (stack,j,SOME list); true)

   fun addSucc ((stack,_), j, a)	= case Stack.sub (stack, j)
       of SOME list =>  Stack.update (stack, j, SOME (a::list))
	| NONE	    =>  Stack.update (stack, j, SOME [a])

   fun existsNext ((stack,table),i,a) = case T.sub (table, (i,a))
       of SOME j        =>      true
	| NONE		=>	false

   fun justNext (forest as (stack,table), i, a) = case T.sub (table, (i,a))
	of SOME j => j
         | NONE   => let
                        val j = Stack.push (stack,NONE)
                        val _ = T.update (table, (i,a), j)
                        val _ = addSucc (forest, i, a)
                        in j
                        end

   fun next (forest as (stack,table), i, a) = case T.sub (table, (i,a))
       of SOME j	=>	(true,j)
	| NONE		=>	let
                        val j = Stack.push (stack,NONE)
			val _ = T.update (table, (i,a), j)
			val _ = addSucc (forest, i, a)
			in (false,j)
			end

	(* ... yields true, if next (i,a) is defined ... *)

   fun succs ((stack,_), j)	= Stack.sub (stack, j)

   fun setSuccs ((stack,_), j, list)	= Stack.update (stack, j, SOME list)

   fun new (stack,_) = Stack.push (stack, NONE)

   fun insert (forest, r, list) = let		(* ok *)

       fun create (j,[]) = if initSucc (forest, j, []) then ()
			else print "Error in creating new node ...\n"

        |  create (j,a::aa) = create (justNext (forest, j, a), aa)

       fun doit (r,[]) = initSucc (forest,r,[])
        |  doit (r, a::aa) = let
	     val (flag, j) = next (forest, r, a)
	     in if flag then doit (j, aa)
                else (create (j, aa);
		     true)
	     end
       in doit (r,list)
       end	(* end of insert *)

		(* ... returns true if list is allowed and new ... *)

   fun hasElem (forest, r, []) = (case succs (forest, r)
	of SOME []	=> true
	|	_	=> false
	)
    |  hasElem (forest, r, x::xs) = if existsNext (forest, r, x)
	  then hasElem (forest, justNext (forest, r, x), xs)
	  else false
(*
   fun toList (forest, r, []) = let
       fun call i = case succs (forest, i)
           of SOME [] => [[]] 
            | SOME list => let
                 fun f a = let
		     val j = justNext (forest, i, a)
		     in map (fn x => a::x) (call j)
		     end
                 in List.concat (map f list)
                 end
            | NONE => []
       in  call r
       end

    |  toList (forest, r, x::xs) = if existsNext (forest, r, x)
	      then let
	         val j = justNext (forest, r, x)
		 in toList (forest, j, xs)
		 end
	      else []
*)
   fun toList (forest, r, []) = let
       fun call i = case succs (forest, i)
           of SOME [] => [[]]
            | SOME list => let
                 fun f (a,acc) = let
                     val j = justNext (forest, i, a)
                     in foldr (fn (x,acc) => (a::x)::acc) acc (call j)
                     end
                 in foldr f [] list
                 end
            | NONE => []
       in  call r
       end

    |  toList (forest, r, x::xs) = if existsNext (forest, r, x)
              then let
                 val j = justNext (forest, r, x)
                 in toList (forest, j, xs)
                 end
              else []

   fun appF f (forest, r, []) = let
       fun call g i = case succs (forest, i)
           of SOME [] => g []
            | SOME list => let
                 fun doit a = let
                     val j = justNext (forest, i, a)
		     fun h l = g (a::l)
                     in call h j
                     end
                 in app doit list
                 end
            | NONE => ()
       in  call f r
       end

    |  appF f (forest, r, x::xs) = if existsNext (forest, r, x)
              then let
                 val j = justNext (forest, r, x)
                 in appF f (forest, j, xs)
                 end
              else ()

   fun foldF f b (forest, r, []) = let
       fun call f (i,acc) = case succs (forest, i)
           of SOME [] => f ([],acc)
            | SOME list => let
                 fun doit (a,acc) = let
                     val j = justNext (forest, i, a)
                     fun g (l,acc) = f (a::l,acc)
                     in call g (j, acc)
                     end
                 in foldr doit acc list
                 end
            | NONE => b
       in  call f (r,b)
       end

    |  foldF f b (forest, r, x::xs) = if existsNext (forest, r, x)
              then let
                 val j = justNext (forest, r, x)
                 in foldF f b (forest, j, xs)
                 end
              else b
(*
   fun count (forest, r) = let
       fun count r = case succs (forest, r)
           of SOME []     =>    1
            | SOME aa     =>    let
                   fun countNext a = count (justNext (forest, r, a))
                   in IntUtils.sum (map countNext aa)
                   end
            | NONE        =>    0
       in count r
       end
*)
   fun count (forest,r) = foldF (fn (l,acc) => 1 + acc) 0 (forest,r,[])

end
