structure ForestMap : ForestMap =

struct
   structure T = Table (structure Item = IntPairItem
			    type value = int )

   type 'a map = 'a option Stack.stack * T.table

   
   fun makeMap n = (Stack.makeStack (n,NONE), T.makeTable n) : 'a map
   fun nullMap () = (Stack.new (), T.nullTable ()) 	     : 'a map

   fun next ((stack,table), i, a) = case T.sub (table, (i,a))
       of SOME j	=>	(false,j)
	| NONE		=>	let
                        val j = Stack.push (stack,NONE)
			val _ = T.update (table, (i,a), j)
			in (true, j)
			end

	(* ... yields false, if next (i,a) is defined ... *)

   fun subNode ((stack,_), j)	  = Stack.sub (stack,j)
   fun updateNode ((stack,_),j,a) = Stack.update (stack,j,SOME a)
   fun addNode ((stack,_),j,a) = case Stack.sub (stack,j)
		of SOME list => Stack.update (stack,j,SOME (a::list))
		|  NONE      => Stack.update (stack,j,SOME [a])

   fun sub (fmap, r, list) = let

       fun doit (r,[]) = subNode (fmap,r)
        |  doit (r, a::aa) = let
	     val (flag, j) = next (fmap, r, a)
	     in if not flag then doit (j, aa)
                else NONE
	     end
       in doit (r,list)
       end	(* end of sub *)

   fun update (fmap,r,list,a) = let

       fun doit (r,[])    = updateNode (fmap,r,a)
	|  doit (r,x::xs) = let
		val (_,j) = next (fmap, r, x)
		in doit (j, xs)
		end

       in  doit (r,list)
       end

   fun add (fmap,r,list,a) = let

       fun doit (r,[])    = addNode (fmap,r,a)
        |  doit (r,x::xs) = let
                val (_,j) = next (fmap, r, x)
                in doit (j, xs)
                end

       in  doit (r,list)
       end

   fun numberOfNodes (stack,_) = Stack.length stack
end
