functor Pool (structure Item: Item ) =

   struct 

(* Buckets ... *)

      type  bucket = Item.item list
      val nullBucket = nil : bucket

      fun hasElemBucket (bucket, item) = 
	 let fun doit nil 	 = false
	      |  doit (a ::rest) = if a=item then true
			           else doit rest
	 in doit bucket
	 end
		
      fun insertBucket (bucket, item) =
	  let fun doit nil = (true,[item])
               |  doit (a::rest) = if a=item then (false, a :: rest)
                           else let
		 	      val (flag, list) = doit rest
	        	      in (flag, a::list)
			      end
         in doit bucket
         end

       fun remBucket (bucket, item) =
         let fun doit [] = []
              |  doit (x::rest) = if x = item then rest
                                        else x ::(doit rest)
         in doit bucket
         end

      fun printBucket  ([]:bucket)    = ()
       |  printBucket  [item] = ( 	print (Item.toString item);
				        print "\n")
      |   printBucket  (item::rest) = (print (Item.toString item);
                                        print ", ";
					printBucket rest)
					  
(* Pools ... *)

      datatype pool = Pool of {buckets  :  bucket array ref,
		      hash 	: (Item.item -> int) ref,
		      count     : int ref  		(* number of entries *)
			  }

      fun hashWord n (x:word) =  Word.toInt (x mod (Word.fromInt n))

      fun makePool n = Pool
	 {buckets    = ref (Array.array(n,nullBucket)),
	  hash    = ref (fn x => hashWord n (Item.toWord x)),
	  count   = ref 0}
      
      fun nullPool() = makePool 2

      fun hasElem (Pool {buckets, hash, count}, item) = let
	  val w = !hash item
	  in hasElemBucket (Array.sub (!buckets, w), item)
	  end

      fun doublePool (Pool {buckets, hash, count}) = let
	  val n = Array.length (!buckets)
          val newN = 2*n

	  val _ = hash  := (fn x => hashWord newN (Item.toWord x))
	  val newBuckets = Array.array (newN, [])

	  fun moveBucket i = let 
	      val bucket = Array.sub (!buckets,i)
	      fun doit []      = ()
	       |  doit (x::xs) = let
		  val j = !hash x
		  val list = Array.sub (newBuckets, j)
		  in 
		     Array.update (newBuckets, j, x::list);
		     doit xs
		  end
	      in doit bucket
              end
          fun doit i = if i >= n then ()
	      else (moveBucket i; doit (i+1))
	  
          in doit 0; buckets := newBuckets
	  end

      fun remElem (pool as Pool {buckets, hash, count}, item) = let
          val i = !hash item
          val bucket = remBucket (Array.sub (!buckets, i), item)
          in Array.update (!buckets, i, bucket)
          end

      fun insertElem (pool as Pool {buckets, hash, count}, item) = let
	  val n = Array.length (!buckets)
	  val _ = if (!count >= n) then doublePool pool
		  else ()
	  val i = !hash item
	  val bucket = Array.sub (!buckets, i)
          val (flag,list) = insertBucket (bucket, item)
	  in
		if flag then (
		   Array.update (!buckets, i, list);
		   count := (!count) + 1; true)
		else false
	  end

(* insertElem returns true if the element is new *)

      fun countInsert (Pool {count,...}) = !count

      fun addElem (Pool {buckets, hash, count}, item) = let
          val i = !hash item
          val bucket = Array.sub (!buckets, i)
          in Array.update (!buckets, i, item :: bucket);
                   count := (!count) + 1
          end

      fun toList (Pool {buckets,...}) = let
          val n = Array.length (!buckets)
	  fun doit (j, result) = if j = 0 then List.concat result
		else doit (j-1, (Array.sub (!buckets,j-1))::result)
	  in doit (n,[])
	  end
	
      fun fromList list = let
	  val pool = makePool (2 * List.length list +1)
          fun f x = (insertElem (pool,x);())
	  val _ = app f list
	  in pool
	  end
	  
      fun copyPool (Pool {buckets, hash, count}) = let
	  val buckets = ref (Array.tabulate (Array.length (!buckets), fn i => 
			Array.sub (!buckets, i)))
	  val hash = ref (!hash)
	  val count = ref (!count)
	  in Pool {buckets = buckets, hash = hash, count = count}
	  end

      fun resetPool (Pool {buckets, hash, count}) = (
		count := 0;
	        buckets := Array.array (Array.length (!buckets),[])
		)
	    
      fun printPool (Pool {buckets,...}) = let
	  val n = Array.length (!buckets)
	  fun doit i = if i >= n then ()
		else (printBucket (Array.sub (!buckets, i));
		      doit (i+1))
	  in  doit 0
	  end
   end
