functor PrintClause (structure Item : Item) =

struct open Horn

   structure S = SymbolTable (structure Item = Pair (
		structure First = StringItem
		structure Second = IntItem))

   structure T = SymbolTable (structure Item = Item)

   fun toString (table, universe, clause) = let

	fun itemToList x = String.explode (
				Item.toString (
				T.getItem (universe,x)))

	fun intToList x  = String.explode (Int.toString x) 

	fun doInd (Const x, acc) = itemToList x :: acc
	 |  doInd (Var x, acc) = (#"_" :: intToList x) :: acc
	
	fun doArgs (args, acc) = let
		fun doit [] = [#")", #" "] :: acc
		 |  doit [a] = doInd (a, [#")"] :: acc)
		 |  doit (a::aa) = doInd (a, [#",", #" "] :: doit aa)
		in [#" ", #"("] :: doit args
		end

	fun doPre (Yes, acc) = [#"1"] :: acc
	 |  doPre (No, acc)  = [#"0"] :: acc
	 |  doPre (U (n,args), acc) = let
		val (name,_) = S.getItem (table,n)
		in String.explode name :: doArgs (args, acc)
		end 
	 |  doPre (N (n,args), acc) = [#"!"] :: doPre (U (n,args), acc)
	 |  doPre (Eq (a,b), acc) = doInd (a, [#" ", #"=", #" "] 
					:: doInd (b, acc))
	 |  doPre (Neq (a,b), acc) = doInd (a, [#" ", #"!", #"=", #" "]
				     	:: doInd (b, acc))
	 |  doPre (And (p1,p2), acc) = [#"("] :: doPre (p1, 
				       [#" ", #"&", #" "] :: doPre (p2,
				       [#")"] :: acc))
	 |  doPre (Or (p1,p2), acc)  = [#"("] :: doPre (p1, 
				       [#" ", #"|", #" "] :: doPre (p2, 
				       [#")"] :: acc))
	 |  doPre (CheapOr (p1,p2), acc)  = [#"("] :: doPre (p1, 
				       [#" ", #"|", #" "] :: doPre (p2, 
				       [#")"] :: acc))
	 |  doPre (Exists (x, p), acc) = [#"(", #"E", #" ", #"_"] :: 
					intToList x :: [#".", #" "] ::
			doPre (p, [#")"] :: acc)
	 |  doPre (CheapExists (x, p), acc) = [#"(", #"E", #" ", #"_"] :: 
                                        intToList x :: [#".", #" "] ::
                        doPre (p, [#")"] :: acc)

	 |  doPre (All (x, p), acc) = [#"(", #"A", #" ", #"_"] :: 
                                        intToList x :: [#".", #" "] ::
                        doPre (p, [#")"] :: acc)

	fun doit (One, acc) = [#"1"] :: acc
	 |  doit (R (n, args), acc) = doPre (U (n,args), acc)
	 |  doit (Implies (p,c), acc) = [#"("] :: doPre (p,
			[#" ", #"=", #">", #" "] :: doit (c, [#")"] :: acc))
	 |  doit (Both (c1,c2), acc) = doit (c1, [#" ", #"&", #"\n"] ::
				doit (c2, acc))
	 |  doit (Forall (x,c), acc) = [#"(", #"A", #" ", #"_"] :: 
					intToList x :: [#".", #" "] ::
					doit (c, [#")"] :: acc)

	in String.implode (List.concat (doit (clause, [[#"\n", #"\n"]])))
	end
end

structure PrintIntClause    = PrintClause (structure Item = IntItem)
structure PrintStringClause = PrintClause (structure Item = StringItem)
structure PrintTreeClause   = PrintClause (structure Item = Tree)

