structure TableData =
struct
datatype ('item,'a) coreTable =       
			      Empty of {zero : 'a}
			|     Single of {content : 'item * 'a,
					 zero : 'a
				}
			|     Small of  {bucket : ('item *'a) list,
					 count : int,
					 zero : 'a
				}
			|     Large of  {buckets  :  ('item*'a) list array,
                      hash      : 'item -> int,
                      count     : int,                  (* number of entries *)
                      zero      : 'a
                          	}
end

functor SimpleTableEq (structure Item: ItemEq) = 

   struct open TableData

      val size = 6

(* Buckets ... *)

      type  'a bucket = (Item.item * 'a) list

      fun searchBucket (bucket, item) = 
	 let fun doit nil = NONE
	      |  doit ((key, value)::rest) = if Item.eq(key,item) 
			then SOME value
		        else doit rest
	 in doit bucket
	 end
		
      fun remBucket (bucket, item) =
	 let fun doit [] = ([],false)
	      |  doit ((key,v)::rest) = if Item.eq(key,item) then (rest,true)
					else let val (result,isIn) = doit rest
					     in ((key,v)::result,isIn)
					     end
	 in doit bucket
	 end
	
      fun extractBucket (bucket, zero, item) =
	 let fun doit [] = (zero,[],false)
	      |  doit ((key,v)::rest) = if Item.eq(key,item) then (v, rest, true)
					else let
				val (u,result,isIn) = doit rest
				in  (u,(key,v):: result,isIn)
				end
	 in doit bucket
	 end
	
      fun updateBucket (bucket : 'a bucket, item, value) =
	  let fun doit nil = (true, [(item,value)])
               |  doit ((a,b)::rest) = if Item.eq(a,item) then 
					  (false, (a,value) :: rest)
                                       else let
	 		val (isNew, list) = doit rest
			in (isNew, (a,b)::list)
			end
         in doit bucket
         end

      fun printBucket f ([] : 'a bucket)    = print "\n"
       |  printBucket f [(item,value)] = (print "(";
					  print (Item.toString item);
					  print ", ";
					  print (f value);
				          print ")\n")
      |   printBucket f ((item,value)::rest) = 
					 (print "(";
                                          print (Item.toString item);
                                          print ", ";
                                          print (f value);
					  print "), ";
					  printBucket f rest)
					  
(* Tables ... *)


      fun getZero (Empty  {zero})     = zero
       |  getZero (Single {zero,...}) = zero			
       |  getZero (Small {zero,...})  = zero
       |  getZero (Large {zero,...})  = zero

      type 'a table = (Item.item,'a) coreTable ref

      
      fun hashWord n (x:word) = Word.toInt (x mod (Word.fromInt n))

      fun nullTable a     = ref (Empty {zero = a}) : 'a table
      fun makeTable (_,a) = nullTable a 

      fun resetTable (table,n) = (table := Empty {zero = getZero (!table)})

      fun sub (ref (Empty{zero}), _)            = zero
       |  sub (ref (Single{content,zero}),item) = let
		val (x,v) = content
		in if Item.eq (item,x) then v
		   else zero
		end
       |  sub (ref (Small {bucket,zero,...}), item) = (case searchBucket (bucket,item)
		of SOME v => v
		|  NONE   => zero
		)
       |  sub (ref (Large {buckets,hash,zero,...}), item) = let
	  val w = hash item
	  in case searchBucket (Array.sub (buckets, w), item)
		of SOME value => value
		|  NONE	      => zero
	  end

      fun extract (ref (Empty {zero}), _)                      = zero
       |  extract (table as ref (Single {content,zero}), item) = let
		val (x,v) = content 
		in if Item.eq (x,item) then ( table := Empty {zero = zero}
					    ; v
					    )
		   else zero
		end
       |  extract (table as ref (Small {bucket,zero,count}), item) = let
		val (v,bucket,isIn) = extractBucket (bucket, zero, item)
		val count = if isIn then count-1
			    else count
		val _ = case bucket
		    of [p] => ( table := Single {content = p, zero=zero} )
		    |  []  => ( table := Empty {zero=zero} )
		    |  _   => ( table := Small  {bucket = bucket,count=count,zero=zero} )  
		in v
		end
       |  extract (table as ref (Large {buckets, hash, count, zero}), item) = let
	  val w = hash item
	  val (v, bucket, isIn) = extractBucket (Array.sub (buckets, w), zero, item)
	  val _ = Array.update (buckets, w, bucket)
	  val count = if isIn then count-1
		      else count
	  val _ = case count 
		   of 0 => let 
				val _ = case Array.foldl (op @) [] buckets
					 of _::_ => raise Fail "Failing empty simple table\n\n"
					 |  _       => ()
			   in ( table := Empty {zero = zero} )
			   end
		   |  1 => let val list = Array.foldl (op @) [] buckets
				val _ = case list
					 of _::_::_ => raise Fail "Failing single simple table\n\n"
					 |  _       => ()
			   in table := Single {content= hd list,zero=zero} 
			   end
		   |  _ => ( table := Large {buckets=buckets, hash=hash, count=count, zero=zero} )
	  in v
	  end

      fun remItem (ref (Empty{zero}), _)                      = ()
       |  remItem (table as ref (Single {content,zero}),item) = let
		val (x,v) = content 
		in if Item.eq (x,item) then table := Empty{zero=zero}
		   else ()
		end
       |  remItem (table as ref (Small {bucket,zero,count}), item) = let
		val (bucket,isIn) = remBucket (bucket, item)
		val count = if isIn then count-1
			    else count
		in case count
		    of 0 => table := Empty {zero=zero}
		    |  1 => table := Single {content = hd bucket, zero = zero}
		    |  _ => table := Small {bucket=bucket,zero=zero,count=count}
		end
       |  remItem (table as ref (Large {buckets, hash, count, zero}), item) = let
	  val w = hash item
	  val (bucket,isIn) = remBucket (Array.sub (buckets, w), item)
	  val _ = Array.update (buckets, w, bucket)
	  val count = if isIn then count-1 
		      else count
	  in case count
                   of 0 => ( table := Empty {zero = zero} )
                   |  1 => let val list = Array.foldl (op @) [] buckets
                           in table := Single {content= hd list,zero=zero}
                           end
                   |  _ => ( table := Large {buckets = buckets,hash=hash,count=count,zero=zero} )
	  end

      fun hasValue (ref (Empty {zero}), _) = false
       |  hasValue (ref (Single {content,...}),item) = let
		val (x,_) = content
		in Item.eq (x,item)
		end
       |  hasValue (ref (Small {bucket,...}),item) = (case searchBucket (bucket,item)
		of SOME _ => true
		|  NONE   => false
		)
       |  hasValue (ref (Large ({buckets, hash,...})), item) = let
	     val w = hash item
             in case searchBucket (Array.sub (buckets, w), item)
                of SOME _ => true
                |  NONE   => false
	     end

      fun doubleTable (buckets : (Item.item * 'a) list array) = let

	  val n    = Array.length buckets
          val newN = 2*n

	  fun hash x     = hashWord newN (Item.toWord x)
	  val newBuckets = Array.array (newN, [])

	  fun moveBucket i = let 
	      val bucket = Array.sub (buckets,i)
	      fun doit []                  = ()
	       |  doit ((index,value)::xs) = let
		  val _    = doit xs
		  val j    = hash index
		  val list = Array.sub (newBuckets, j)
		  in 
		     Array.update (newBuckets, j, (index,value)::list)
		  end
	      in doit bucket
              end
          fun doit i = if i >= n then ()
	      else (moveBucket i; doit (i+1))
	  
          val _ = doit 0
	  in (newBuckets, hash)
	  end

      fun add (table as ref (Empty {zero}), item, value) = 
		( table := Single {content = (item,value), zero=zero} )
       |  add (table as ref (Single {content,zero}), item, value) = 
		( table := Small {bucket = [(item,value),content], zero=zero,count=2} )
       |  add (table as ref (Small {bucket,zero,count}), item, value) = if count < size
		then ( table := Small {bucket = (item,value)::bucket, zero=zero, count = count+1} )
		else let
		   val buckets = Array.array (2*size,[])
		   fun hash x = hashWord (2*size) (Item.toWord x)
 		   fun doit []      = ()
               	    |  doit ((index,value)::xs) = let
                  	val _ = doit xs
                  	val j = hash index
                  	val list = Array.sub (buckets, j)
                  	in Array.update (buckets, j, (index,value)::list)
                  	end
                   val _ = doit ((item,value) :: bucket)
		   in ( table := Large {buckets=buckets,hash=hash,zero=zero,count=count+1} )
		   end
       |  add (table as ref (Large (record as {buckets, hash, count, zero})), item, value) = let
	  val n = Array.length buckets
 	  val (buckets,hash) = 	if (count > (n * 3) div 4) then doubleTable buckets
                  		else (buckets,hash)
	  val w = hash item
          val count = count+1
	  val bucket = (item,value) :: (Array.sub (buckets, w))
       	  val _ =  Array.update (buckets, w, bucket)
	  in (table := Large {buckets=buckets, count=count, hash=hash, zero=zero})
          end

      fun update (table as ref (Empty {zero}), item, value) =
                ( table := Single {content = (item,value), zero=zero} )
       |  update (table as ref (Single {content,zero}), item, value) = let
		  val (x,v) = content
		  in if Item.eq (x,item) then ( table := Single {content = (x,value),zero=zero} )
		     else
                        ( table := Small {bucket = [(item,value),content], zero=zero,count=2} )
		  end
       |  update (table as ref (Small {bucket,zero,count}), item, value) = let
		val (isNew, bucket) = updateBucket (bucket, item, value) 
		val count = if isNew then count+1
			    else count
		in if count < size
                   then ( table := Small {bucket = bucket, zero = zero, count = count} )
                   else let
                      val buckets = Array.array (2*size,[])
                      fun hash x = hashWord (2*size) (Item.toWord x)
                      fun doit []      = ()
                       |  doit ((index,value)::xs) = let
                           val _ = doit xs
                           val j = hash index
                           val list = Array.sub (buckets, j)
                           in Array.update (buckets, j, (index,value)::list)
                           end
                      val _ = doit bucket
                      in ( table := Large {buckets=buckets,hash=hash,zero=zero,count=count} )
                      end
		end
       |  update (table as ref (Large (record as {buckets, hash, count, zero})), item, value) = let
             val w              = hash item
             val (isNew,bucket) = updateBucket (Array.sub (buckets, w), item, value)
             val _              =  Array.update (buckets, w, bucket)
             val count          = if isNew then count+1
			          else count
             val n = Array.length buckets
             val (buckets,hash) =  if (count >= (n * 3) div 4) then doubleTable buckets
                                else (buckets,hash)
             in (table := Large {buckets=buckets, count=count, hash=hash, zero=zero})
             end

      fun fromList zero []      = ref (Empty {zero=zero})
       |  fromList zero [(x,v)] = ref (Single {content = (x,v), zero = zero})
       |  fromList zero (list : (Item.item * 'a) list) = let
		val count = List.length list
		in if count <= size then 
			ref (Small {bucket = list, count = count, zero = zero})
		   else let
			val N       = 2 *count
			val buckets = Array.array (N,[])
			fun hash x  = hashWord N (Item.toWord x)
			fun f (x,v) = let
				val j      = hash x
				val bucket = Array.sub (buckets,j)
				in Array.update (buckets,j, (x,v) :: bucket)
				end
			val _ = app f list
			in ref (Large {buckets=buckets, count=count, hash=hash, zero=zero})
			end
		end

      fun toList (ref (Empty _))              = []
       |  toList (ref (Single {content,...})) = [content]
       |  toList (ref (Small {bucket,...}))   = bucket
       |  toList (ref (Large {buckets,...}))  = Array.foldl (op @) [] (buckets)

      fun nullValue (ref table) = getZero table

      fun copyTable (ref (Large {buckets,hash,count,zero})) = let
		val n = Array.length buckets
		val buckets = Array.tabulate (n,fn i => Array.sub (buckets,i))
		in ref (Large {buckets=buckets,hash=hash,count=count,zero=zero})
		end
       |  copyTable (ref z) = ref z

      fun map f table = let
	val zero = nullValue table
	val list = List.map (fn (x,a) => (x,f a)) 
		       (toList table)
	in fromList (f zero) list 
	end

      fun map0 (zero,f) table = let
	val list = List.map (fn (x,a) => (x,f (x,a)))
		       (toList table)
	in fromList zero list 
	end

(* Statistics .... *)

      fun max (ref (Empty _))             = 0
       |  max (ref (Single _))            = 1
       |  max (ref (Small {count,...}))   = count
       |  max (ref (Large {buckets,...})) = let
	  val n = Array.length buckets
	  fun doit (j,r) = if j >=n then r
		else doit (j+1, Int.max (r, 
			   List.length (Array.sub (buckets, j))))
	  in  if n = 0 then 0
	      else doit (1, List.length (Array.sub (buckets, 0)))
	  end

      fun zeroFraction (ref (Empty _))  = 0.0
       |  zeroFraction (ref (Single _)) = 0.0
       |  zeroFraction (ref (Small _))  = 0.0
       |  zeroFraction (ref (Large {buckets,count,...})) = let
	  val n = Array.length (buckets)
	  fun doit (j,r) = if j >=n then r
		else let val l = List.length (Array.sub (buckets, j)) 
		     in if l > 0 then doit (j+1,r)
			else doit (j+1,r+1)
		     end
	  val zeros = Real.fromInt (doit (0,0))
	  in  zeros / (Real.fromInt n)
	  end

      fun average (ref (Small {count,...})) = Real.fromInt count
       |  average (ref (Single _))          = 1.0
       |  average (ref (Empty _))           = 0.0
       |  average (ref (Large {buckets,count,...})) = let
	  val length = Array.length buckets
	  fun doit (j,n,s) = if j >=length then (n,s)
		else let val l = List.length (Array.sub (buckets, j)) 
		     in if l > 0 then doit (j+1,n+1, s+l)
			else doit (j+1,n,s)
		     end
	  val (n,s) = doit (0,0,0)
	  in  (Real.fromInt s) / (Real.fromInt n)
	  end

      fun printTable f (ref (Empty _)) = print "\n"
       |  printTable f (ref (Single {content=(item,value),...})) = 
					( print "("
                                        ;  print (Item.toString item)
                                        ;  print ", "
                                        ;  print (f value)
                                        ;  print ")\n"
					)
       |  printTable f (ref (Small {bucket,...})) = printBucket f bucket
       |  printTable f (ref (Large {buckets,...})) = let
	     val n = Array.length buckets
	     fun doit i = if i >= n then ()
		   else ( printBucket f (Array.sub (buckets, i))
		        ; doit (i+1)
			)
	     in  doit 0
	     end
   end
