signature Stack =
sig
   type 'a stack
   val new : 	unit -> 'a stack
   val makeStack : int * 'a -> 'a stack
   val isEmpty: 'a stack -> bool
   val push :	'a stack * 'a -> int
   val pop :	'a stack -> 'a option
   val sub :	'a stack * int -> 'a
   val update :	'a stack * int * 'a -> unit
   val length : 'a stack -> int
   val toList : 'a stack -> 'a list
   val fromList: 'a list -> 'a stack
   val toString: ('a -> string) -> 'a stack -> string
(*   val resetStack : 'a stack * int * 'a -> unit *)
   val resetStack: 'a stack -> unit


end

structure Stack :> Stack =
struct
   type 'a stack = (int * 'a Array.array) ref

   fun new () = ref (0,Array.fromList [])

   fun makeStack (n, a) = ref (n, Array.array (n,a))

   fun isEmpty (ref (count,_))  = if count = 0 then true else false

   fun length (ref (n,_)) = n

   fun push (stack as ref (count,array), a) = let
       	  val length = Array.length array
          in if count < length then 
	         (Array.update (array,count,a); 
		  stack := (count+1, array); count)
              else let
                 fun copy i = if i < length then
                      Array.sub (array,i)
                      else a
                 val newLength = if length = 0 then 1 
				 else 2 * length
                 val newArray = Array.tabulate (newLength, copy)
                 in stack := (count+1, newArray); count
                 end
          end
    
   fun pop stack = if isEmpty stack then NONE
		   else let
		      val ref (n, array) = stack
		      val return = Array.sub (array,n-1)
		      val _ = stack := (n-1, array) 
		      in SOME return 
                      end

  (* fun resetStack (stack,n,a) = stack := !(makeStack(n,a)) *)
   fun resetStack (stack) = stack := !(new())

   fun sub (ref (_,array),i)     = Array.sub (array,i)

   fun update (ref (_,array),i,a) = Array.update (array,i,a)

   fun fromList list = let
       val array = Array.fromList (List.rev list)
       val n = Array.length array
       in
	   ref (n,array)
       end

   fun toList (ref (n,array)) = let
       fun doit (m, list) = if m >= n then list
		else let
		   val a = Array.sub (array,m)
                in doit (m+1,a::list)
                end
       in doit (0, [])
       end

   fun toString f = String.concat o (map (fn i => (" "^(f i)))) o toList
end

   
	     
