functor Table (structure Item: Item 
	       type value) =

   struct

(* Buckets ... *)

      type  bucket = (Item.item * value) list
      val nullBucket = nil : bucket

      fun searchBucket (bucket, item) = 
	 let fun doit nil = NONE
	      |  doit ((key, value)::rest) = if key=item then SOME value
					     else doit rest
	 in doit bucket
	 end
		
      fun remBucket (bucket, item) =
	 let fun doit [] = []
	      |  doit ((key,v)::rest) = if key = item then rest
					else (key,v)::(doit rest)
	 in doit bucket
	 end
	
      fun updateBucket (bucket : bucket, item, value) =
	  let fun doit nil = (true, [(item,value)])
               |  doit ((a,b)::rest) = if a=item then 
					  (false, (a,value) :: rest)
                                       else let
	 		val (flag, list) = doit rest
			in (flag, (a,b)::list)
			end
         in doit bucket
         end

      fun modifyBucket f bucket = let
	  fun g (item,value) = (item,f value)
	  in map g bucket
	  end

      fun modifyiBucket f bucket = let
	  fun g (item,value) = (item, f (item,value))
	  in map g bucket
 	  end

      fun printBucket f ([]:bucket)    = print "\n"
       |  printBucket f [(item,value)] = (print "(";
					  print (Item.toString item);
					  print ", ";
					  print (f value);
				          print ")\n")
      |   printBucket f ((item,value)::rest) = 
					 (print "(";
                                          print (Item.toString item);
                                          print ", ";
                                          print (f value);
					  print "), ";
					  printBucket f rest)
					  
(* Tables ... *)

      type table = {buckets  :  bucket array ref,
		      hash 	: (Item.item -> int) ref,
		      count     : int ref  		(* number of entries *)
			  }

      fun hashWord n (x:word) = Word.toInt (x mod (Word.fromInt n))

      fun makeTable n = 
	 {buckets    = ref (Array.array(n,nullBucket)),
	  hash    = ref (fn x => hashWord n (Item.toWord x)),
	  count   = ref 0}
      
      fun nullTable() = makeTable 2

      fun resetTable ({buckets,hash,count},n) = 
		( buckets :=  Array.array(n,nullBucket)
		; hash    := (fn x => hashWord n (Item.toWord x))
		; count	  := 0
		)

      fun sub ({buckets, hash, count}:table, item) = let
	  val w = !hash item
	  in searchBucket (Array.sub (!buckets, w), item)
	  end

      fun remItem ({buckets, hash, count}:table, item) = let
	  val w = !hash item
	  val bucket = remBucket (Array.sub (!buckets, w), item)
	  in Array.update (!buckets, w, bucket)
	  end

      fun hasValue (table, item) = case (sub (table, item))
	  of SOME _ => true
	   | NONE   => false

      fun doubleTable ({buckets, hash, count}: table) = let
	  val n = Array.length (!buckets)
          val newN = 2*n

	  val _ = hash  := (fn x => hashWord newN (Item.toWord x))
	  val newBuckets = Array.array (newN, [])

	  fun moveBucket i = let 
	      val bucket = Array.sub (!buckets,i)
	      fun doit []      = ()
	       |  doit ((index,value)::xs) = let
		  val _ = doit xs
		  val j = !hash index
		  val list = Array.sub (newBuckets, j)
		  in 
		     Array.update (newBuckets, j, (index,value)::list)
		  end
	      in doit bucket
              end
          fun doit i = if i >= n then ()
	      else (moveBucket i; doit (i+1))
	  
          in doit 0; buckets := newBuckets
	  end

      fun add (table as {buckets, hash, count}:table, item, value) = let
	  val n = Array.length (!buckets)
 	  val _ = if (!count >= (n * 1) div 3) then doubleTable table
                  else ()
	  val w = !hash item
          val _ = (count := !count +1)
	  val bucket = (item,value) :: (Array.sub (!buckets, w))
       	  in  Array.update (!buckets, w, bucket)
          end

      fun update (table as {buckets, hash, count}: table, item, value) = let
	  val n = Array.length (!buckets)
	  val _ = if (!count >= (n * 1) div 3) then doubleTable table
		  else ()
	  val i = !hash item
	  val bucket = Array.sub (!buckets, i)
	  val (flag, bucket) = updateBucket (bucket, item, value)
	  val _ =  Array.update (!buckets, i, bucket)
	  in if flag then count := (!count) + 1
	     else ()
	  end

      fun fromList list = let
	  val table = makeTable (2 * List.length list +1)
	  fun doit [] = table
	   |  doit ((item,value) :: rest) = (add (table, item ,value); doit rest)
	  in  doit list
	  end

      fun toList ({buckets,...}:table) = List.concat (Array.foldl (op ::) [] (!buckets))

     fun copyTable (table as {buckets, hash, count} : table) = let
          val buckets = ref (Array.tabulate (Array.length (!buckets),
                        fn i => Array.sub (!buckets,i)))
          val hash = ref (!hash)
          val count = ref (!count)
          in {buckets = buckets,hash = hash,count = count}
          end

     fun modifyTable f (table as {buckets, hash, count} : table) =
	 Array.modify (modifyBucket f) (!buckets)

     fun modifyiTable f (table as {buckets, hash, count} : table) =
	 Array.modify (modifyiBucket f) (!buckets)

(*
      val copyTable = fromList o toList
*)

(* Statistics .... *)

      fun max ({buckets,...}:table) = let
	  val n = Array.length (!buckets)
	  fun doit (j,r) = if j >=n then r
		else doit (j+1, Int.max (r, 
			   List.length (Array.sub (!buckets, j))))
	  in  if n = 0 then 0
	      else doit (1, List.length (Array.sub (!buckets, 0)))
	  end

      fun zeroFraction ({buckets,count,...}:table) = let
	  val n = Array.length (!buckets)
	  fun doit (j,r) = if j >=n then r
		else let val l = List.length (Array.sub (!buckets, j)) 
		     in if l > 0 then doit (j+1,r)
			else doit (j+1,r+1)
		     end
	  val zeros = Real.fromInt (doit (0,0))
	  in  zeros / (Real.fromInt n)
	  end

      fun average ({buckets,count,...}:table) = let
	  val length = Array.length (!buckets)
	  fun doit (j,n,s) = if j >=length then (n,s)
		else let val l = List.length (Array.sub (!buckets, j)) 
		     in if l > 0 then doit (j+1,n+1, s+l)
			else doit (j+1,n,s)
		     end
	  val (n,s) = doit (0,0,0)
	  in  (Real.fromInt s) / (Real.fromInt n)
	  end

      fun printTable f ({buckets,...}: table) = let
	  val n = Array.length (!buckets)
	  fun doit i = if i >= n then ()
		else (printBucket f (Array.sub (!buckets, i));
		      doit (i+1))
	  in  doit 0
	  end

(*
      structure T = TableData 

      fun fromSimpleTable (ref (T.Empty _)) = nullTable () 
       |  fromSimpleTable (ref (T.Single {content,...})) = fromList [content]
       |  fromSimpleTable (ref (T.Small {bucket,...})) = fromList bucket
       |  fromSimpleTable (ref (T.Large {buckets,hash,count,zero : value})) =
			{buckets = ref buckets, hash = ref hash, count = ref count}
*)
 
   structure T = SimpleTable (structure Item = Item)
   fun fromSimpleTable table = fromList (T.toList table)

   end
